__author__ = "Simon Nilsson"

import os
from statistics import mean
from typing import List, Optional

import numpy as np
import pandas as pd

from simba.feature_extractors.perimeter_jit import jitted_centroid
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class MovementCalculator(ConfigReader, FeatureExtractionMixin):
    """
    Compute aggregate movement statistics from pose-estimation data in SimBA project.

    :parameters str config_path: path to SimBA project config file in Configparser format
    :param List[str] body_parts: Body-parts to use for movement calculations OR ``Animal_name CENTER OF GRAVITY``. If ``Animal_name CENTER OF GRAVITY``, then SimBA will approximate animal centroids through convex hull.
    :param float threshold: Filter body-part detection below set threshold (Value 0-1). Default: 0.00
    :param List[str] or None file_paths: Files to calucalte movements for. If None, then all files in ``project_folder/csv/outlier_corrected_movement_location`` directory.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.

    :examples:
    >>> body_parts=['Animal_1 CENTER OF GRAVITY']
    >>> movement_processor = MovementCalculator(config_path='project_folder/project_config.ini', body_parts=body_parts)
    >>> movement_processor.run()
    >>> movement_processor.save()

    """

    def __init__(
        self,
        config_path: str,
        body_parts: List[str],
        threshold: float = 0.00,
        file_paths: Optional[List[str]] = None,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        FeatureExtractionMixin.__init__(self)
        self.save_path = os.path.join(
            self.logs_path, "Movement_log_{}.csv".format(self.datetime)
        )
        self.file_paths, self.body_parts, self.threshold = (
            file_paths,
            body_parts,
            threshold,
        )
        if not self.file_paths:
            check_if_filepath_list_is_empty(
                filepaths=self.outlier_corrected_paths,
                error_msg=f"SIMBA ERROR: Cannot process movement. ZERO data files found in the {self.outlier_corrected_dir} directory.",
            )
            self.file_paths = self.outlier_corrected_paths
        print(f"Processing {len(self.file_paths)} video(s)...")

    def __find_body_part_columns(self):
        self.body_parts_dict, self.bp_list = {}, []
        for bp_cnt, bp_name in enumerate(self.body_parts):
            if not bp_name.endswith("CENTER OF GRAVITY"):
                animal_name = self.find_animal_name_from_body_part_name(
                    bp_name=bp_name, bp_dict=self.animal_bp_dict
                )
                self.body_parts_dict[bp_cnt] = {
                    "ANIMAL NAME": animal_name,
                    "BODY-PART": bp_name,
                    "BODY-PART HEADERS": [
                        f"{bp_name}_x",
                        f"{bp_name}_y",
                        f"{bp_name}_p",
                    ],
                }
                self.bp_list.extend((self.body_parts_dict[bp_cnt]["BODY-PART HEADERS"]))
            else:
                pass

    def __find_polygons(self, data):
        print(data.shape)

    def run(self):
        self.results = pd.DataFrame(
            columns=["VIDEO", "ANIMAL", "BODY-PART", "MEASURE", "VALUE"]
        )
        self.movement_dfs = {}
        for file_path in self.file_paths:
            self.__find_body_part_columns()
            _, video_name, _ = get_fn_ext(file_path)
            print("Analysing {}...".format(video_name))
            self.data_df = read_df(file_path=file_path, file_type=self.file_type)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=video_name
            )
            self.movement_dfs[video_name] = pd.DataFrame()
            if self.bp_list:
                self.data_df = self.data_df[self.bp_list]
                for animal_cnt, animal_data in self.body_parts_dict.items():
                    animal_df = self.data_df[animal_data["BODY-PART HEADERS"]]
                    if self.threshold > 0.00:
                        animal_df = animal_df[
                            animal_df[animal_data["BODY-PART HEADERS"][-1]]
                            >= self.threshold
                        ]
                    animal_df = animal_df.iloc[:, 0:2].reset_index(drop=True)
                    animal_df = self.create_shifted_df(df=animal_df)
                    bp_time_1 = animal_df[
                        [
                            animal_data["BODY-PART HEADERS"][0],
                            animal_data["BODY-PART HEADERS"][1],
                        ]
                    ].values.astype(float)
                    bp_time_2 = animal_df[
                        [
                            animal_data["BODY-PART HEADERS"][0] + "_shifted",
                            animal_data["BODY-PART HEADERS"][1] + "_shifted",
                        ]
                    ].values.astype(float)
                    self.movement = pd.Series(
                        self.framewise_euclidean_distance(
                            location_1=bp_time_1,
                            location_2=bp_time_2,
                            px_per_mm=self.px_per_mm,
                        )
                    )
                    self.movement.loc[0] = 0
                    self.movement_dfs[video_name][
                        f'{animal_data["ANIMAL NAME"]} {animal_data["BODY-PART"]}'
                    ] = self.movement
                    distance = round((self.movement.sum() / 10), 4)
                    velocity_lst = []
                    for df in np.array_split(
                        self.movement, int(len(self.movement) / self.fps)
                    ):
                        velocity_lst.append(df.sum())
                    self.results.loc[len(self.results)] = [
                        video_name,
                        animal_data["ANIMAL NAME"],
                        animal_data["BODY-PART"],
                        "Distance (cm)",
                        distance,
                    ]
                    self.results.loc[len(self.results)] = [
                        video_name,
                        animal_data["ANIMAL NAME"],
                        animal_data["BODY-PART"],
                        "Velocity (cm/s)",
                        round((mean(velocity_lst) / 10), 4),
                    ]

            else:
                for animal in self.body_parts:
                    animal_name = animal.split("CENTER OF GRAVITY")[0].strip()
                    x, y = (
                        self.data_df[self.animal_bp_dict[animal_name]["X_bps"]],
                        self.data_df[self.animal_bp_dict[animal_name]["Y_bps"]],
                    )
                    z = pd.concat([x, y], axis=1)[
                        [item for items in zip(x.columns, y.columns) for item in items]
                    ]
                    df = pd.DataFrame(
                        jitted_centroid(
                            points=np.reshape(z.values, (len(z / 2), -1, 2)).astype(
                                np.float32
                            )
                        ),
                        columns=["X", "Y"],
                    )
                    df = self.dataframe_savgol_smoother(df=df, fps=self.fps).astype(int)
                    df_shifted = df.shift(1)
                    df_shifted = df_shifted.combine_first(df).add_suffix("_shifted")
                    self.movement = pd.Series(
                        self.framewise_euclidean_distance(
                            location_1=df.values.astype(np.float32),
                            location_2=df_shifted.values.astype(np.float32),
                            px_per_mm=self.px_per_mm,
                        )
                    )
                    self.movement.loc[0] = 0
                    self.movement_dfs[video_name][
                        f'{animal_name} {"GRAVITY CENTER"}'
                    ] = self.movement
                    distance = round((self.movement.sum() / 10), 4)
                    velocity_lst = []
                    for df in np.array_split(
                        self.movement, int(len(self.movement) / self.fps)
                    ):
                        velocity_lst.append(df.sum())
                    self.results.loc[len(self.results)] = [
                        video_name,
                        animal_name,
                        "GRAVITY CENTER",
                        "Distance (cm)",
                        distance,
                    ]
                    self.results.loc[len(self.results)] = [
                        video_name,
                        animal_name,
                        "GRAVITY CENTER",
                        "Velocity (cm/s)",
                        round((mean(velocity_lst) / 10), 4),
                    ]

    def save(self):
        self.results.set_index("VIDEO").to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Movement log saved in {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = MovementCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                           body_parts=['Simon CENTER OF GRAVITY'], #['Simon CENTER OF GRAVITY', 'JJ CENTER OF GRAVITY']
#                           threshold=0.00)
# test.run()


# test.save()

# test = MovementCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini',
#                           body_parts=['Animal_1 CENTER OF GRAVITY'], #['Simon CENTER OF GRAVITY', 'JJ CENTER OF GRAVITY']
#                           threshold=0.00)
# test.run()
# test.save()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.pose_processors.remove_keypoints import KeypointRemover
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FolderSelect)
from simba.utils.enums import Formats, Keys, Links


class DropTrackingDataPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(
            self, title="Drop body-parts in pose-estimation data", size=(500, 800)
        )
        file_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="FILE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_TOOLS.value,
        )
        self.data_folder_path = FolderSelect(
            file_settings_frm, "Data Folder", lblwidth="20"
        )
        self.file_format = DropDownMenu(
            file_settings_frm, "File Type", ["csv", "h5"], "20"
        )
        self.pose_tool = DropDownMenu(
            file_settings_frm, "Tracking tool", ["DLC", "maDLC"], "20"
        )
        self.pose_tool.setChoices("DLC")
        self.file_format.setChoices("csv")
        self.bp_cnt = Entry_Box(
            file_settings_frm, "# body-parts to remove", "20", validation="numeric"
        )
        confirm_btn = Button(
            file_settings_frm, text="Confirm", command=lambda: self.confirm()
        )
        file_settings_frm.grid(row=0, sticky=NW)
        self.data_folder_path.grid(row=0, sticky=W, columnspan=3)
        self.pose_tool.grid(row=1, sticky=NW)
        self.file_format.grid(row=2, sticky=NW)
        self.bp_cnt.grid(row=3, sticky=NW)
        confirm_btn.grid(row=3, column=1, sticky=NW)

    def confirm(self):
        if hasattr(self, "bp_table"):
            self.bp_table.destroy()
        self.keypoint_remover = KeypointRemover(
            data_folder=self.data_folder_path.folder_path,
            pose_tool=self.pose_tool.getChoices(),
            file_format=self.file_format.getChoices(),
        )
        self.bp_table = LabelFrame(
            self.main_frm,
            text="REMOVE BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.bp_table.grid(row=1, sticky=NW, pady=5)
        self.animal_names_lst, self.drop_down_list = [], []
        if self.pose_tool.getChoices() == "DLC":
            for bp_number in range(int(self.bp_cnt.entry_get)):
                bp_drop_down = DropDownMenu(
                    self.bp_table,
                    "Body-part {}:".format(str(bp_number + 1)),
                    self.keypoint_remover.body_part_names,
                    "10",
                )
                bp_drop_down.setChoices(
                    self.keypoint_remover.body_part_names[bp_number]
                )
                self.drop_down_list.append(bp_drop_down)
                bp_drop_down.grid(row=bp_number, column=0, sticky=NW)
        if self.pose_tool.getChoices() == "maDLC":
            for bp_number in range(int(self.bp_cnt.entry_get)):
                animal_drop_down = DropDownMenu(
                    self.bp_table,
                    "Animal name:",
                    self.keypoint_remover.animal_names,
                    "10",
                )
                animal_drop_down.setChoices(self.keypoint_remover.animal_names[0])
                self.animal_names_lst.append(animal_drop_down)
                bp_drop_down = DropDownMenu(
                    self.bp_table,
                    "Body-part {}:".format(str(bp_number + 1)),
                    self.keypoint_remover.body_part_names,
                    "10",
                )
                bp_drop_down.setChoices(
                    self.keypoint_remover.body_part_names[bp_number]
                )
                self.drop_down_list.append(bp_drop_down)
                animal_drop_down.grid(row=bp_number, column=0, sticky=NW)
                bp_drop_down.grid(row=bp_number, column=1, sticky=NW)

        run_btn = Button(
            self.main_frm, text="RUN BODY-PART REMOVAL", command=lambda: self.run()
        )
        run_btn.grid(row=2, column=0, sticky=NW)

    def run(self):
        bp_to_remove_list, animal_names_list = [], []
        for number, drop_down in enumerate(self.drop_down_list):
            bp_to_remove_list.append(drop_down.getChoices())
        if self.pose_tool == "maDLC":
            for number, drop_down in enumerate(self.animal_names_lst):
                animal_names_list.append(drop_down.getChoices())
        _ = self.keypoint_remover.run(
            bp_to_remove_list=bp_to_remove_list, animal_names=animal_names_list
        )



__author__ = "Simon Nilsson"

import glob
import os
from typing import List, Union

import pandas as pd

from simba.cue_light_tools.cue_light_tools import find_frames_when_cue_light_on
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_that_column_exist
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_config_entry, read_df


class CueLightClfAnalyzer(ConfigReader):
    """
    Compute aggregate statistics when classified behaviors are occurring in relation to the cue light
    ON and OFF states.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int pre_window: Time period (in millisecond) before the onset of each cue light to compute aggregate classification
        statistics within.
    :parameter int post_window: Time period (in millisecond) after the offset of each cue light to compute aggregate classification
        statistics within.
    :parameter List[str] cue_light_names: Names of cue lights, as defined in the SimBA ROI interface.
    :parameter List[str] list: Names of the classifiers we want to compute aggregate statistics for.

    .. note::
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.

    Examples
    ----------
    >>> cue_light_clf_analyzer = CueLightClfAnalyzer(config_path='MyProjectConfig', pre_window=1000, post_window=1000, cue_light_names=['Cue_light'], clf_list=['Attack'])
    >>> cue_light_clf_analyzer.analyze_clf()
    >>> cue_light_clf_analyzer.organize_results()
    >>> cue_light_clf_analyzer.save_data()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        pre_window: int,
        post_window: int,
        cue_light_names: List[str],
        clf_list: List[str],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        self.pre_window, self.post_window = pre_window, post_window
        self.cue_light_names = cue_light_names
        self.clf_list = clf_list
        self.project_path = read_config_entry(
            self.config, "General settings", "project_path", data_type="folder_path"
        )
        self.cue_light_data_dir = os.path.join(self.project_path, "csv", "cue_lights")
        self.machine_results_dir = os.path.join(
            self.project_path, "csv", "machine_results"
        )
        self.no_animals = read_config_entry(
            self.config, "General settings", "animal_no", "int"
        )
        self.files_found_cue_light = glob.glob(
            self.cue_light_data_dir + "/*." + self.file_type
        )
        if len(self.files_found_cue_light) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: No cue light data found. Please analyze cue light data before analyzing classifications based on cue light data"
            )

    def analyze_clf(self):
        """
        Method to calculate classifier data during cue lights periods

        Returns
        -------
        Attribute: dict
            results
        """

        self.results = {}
        self.clf_frms = {}
        for file_cnt, file_path in enumerate(self.files_found_cue_light):
            _, self.video_name, _ = get_fn_ext(file_path)
            machine_results_path = os.path.join(
                self.machine_results_dir, self.video_name + "." + self.file_type
            )
            if not os.path.isfile(machine_results_path):
                print(
                    "SIMBA ERROR: No machine classifications exist for {}."
                    " Skipping cue light classifier analysis for video {}".format(
                        self.video_name, self.video_name
                    )
                )
                continue
            else:
                self.results[self.video_name] = {}
                cue_light_df = read_df(file_path, self.file_type)
                (
                    self.video_info_settings,
                    self.px_per_mm,
                    self.fps,
                ) = self.read_video_info(video_name=self.video_name)
                self.prior_window_frames_cnt = int(self.pre_window / (1000 / self.fps))
                self.post_window_frames_cnt = int(self.post_window / (1000 / self.fps))
                machine_results_df = read_df(machine_results_path, self.file_type)
                data_df = pd.concat(
                    [machine_results_df, cue_light_df[self.cue_light_names]], axis=1
                )
                del cue_light_df, machine_results_df
                cue_light_frames_dict = find_frames_when_cue_light_on(
                    data_df=data_df,
                    cue_light_names=self.cue_light_names,
                    fps=self.fps,
                    prior_window_frames_cnt=self.prior_window_frames_cnt,
                    post_window_frames_cnt=self.post_window_frames_cnt,
                )
                for clf in self.clf_list:
                    check_that_column_exist(
                        df=data_df, column_name=clf, file_name=file_path
                    )
                    self.results[self.video_name][clf] = {}
                    self.clf_frms[clf] = list(data_df.index[data_df[clf] == 1])
                    for cue_light_name, cue_light_data in cue_light_frames_dict.items():
                        self.results[self.video_name][clf][cue_light_name] = {}
                        for period_name, period in zip(
                            ["pre-cue", "cue", "post-cue"],
                            [
                                "pre_window_frames",
                                "light_on_frames",
                                "post_window_frames",
                            ],
                        ):
                            clf_in_period_frms = list(
                                set(self.clf_frms[clf]).intersection(
                                    cue_light_data[period]
                                )
                            )
                            self.results[self.video_name][clf][cue_light_name][
                                period_name
                            ] = round(
                                ((len(clf_in_period_frms) * (1000 / self.fps)) / 1000),
                                4,
                            )

    def organize_results(self):
        """
        Method to organize classifier data into a summary dataframe.

        Returns
        -------
        Attribute: pd.DataFrame
            results_df
        """

        self.results_df = pd.DataFrame(
            columns=["Video", "Classifier", "Cue light", "Time period", "Time (s)"]
        )
        for video_name, video_data in self.results.items():
            for clf_name, clf_data in video_data.items():
                for light_name, light_data in clf_data.items():
                    for period_name, period_data in light_data.items():
                        self.results_df.loc[len(self.results_df)] = [
                            video_name,
                            clf_name,
                            light_name,
                            period_name,
                            period_data,
                        ]
            clf_outside_windows = (
                (len(self.clf_frms[clf_name]) * (1000 / self.fps)) / 1000
            ) - self.results_df["Time (s)"][
                self.results_df["Video"] == video_name
            ].sum()
            if clf_outside_windows < 0.00:
                clf_outside_windows = 0.00
            self.results_df.loc[len(self.results_df)] = [
                video_name,
                clf_name,
                None,
                "Outside light cue time periods",
                clf_outside_windows,
            ]

    def save_data(self):
        """
        Method to save cue light classification data to disk. Results are stored in the `project_folder/logs`
        directory of the SimBA project.

        Returns
        -------
        None
        """

        save_results_path = os.path.join(
            self.logs_path, "Cue_lights_clf_statistics_{}.csv".format(self.datetime)
        )
        self.results_df.to_csv(save_results_path)
        stdout_success(
            msg="SIMBA COMPLETE: Cue light classifier statistics saved in project_folder/logs directory."
        )


# test = CueLightClfAnalyzer(config_path='/Users/simon/Desktop/troubleshooting/light_analyzer/project_folder/project_config.ini',
#                            pre_window=1000,
#                            post_window=1000,
#                            cue_light_names=['Cue_light'],
#                            clf_list=['Attack'])
# test.analyze_clf()
# test.organize_results()
# test.save_data()



__author__ = "Simon Nilsson"

import glob
import os
import re
import struct
import sys
from multiprocessing import Lock, Process, Value
from typing import List

import pandas as pd
from matplotlib import cm

import simba
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.enums import Methods, Paths
from simba.utils.read_write import get_fn_ext


class SharedCounter(object):
    def __init__(self, initval=0):
        self.val = Value("i", initval)
        self.lock = Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def value(self):
        with self.lock:
            return self.val.value


def get_body_part_configurations() -> dict:
    """
    Return dict with named body-part schematics of pose-estimation schemas in SimBA installation as keys,
    and paths to the images representing those body-part schematics as values.
    """

    lookup = {}
    simba_dir = os.path.dirname(simba.__file__)
    img_dir = os.path.join(simba_dir, Paths.SCHEMATICS.value)
    names_path = os.path.join(simba_dir, Paths.PROJECT_POSE_CONFIG_NAMES.value)
    check_file_exist_and_readable(file_path=names_path)
    check_if_dir_exists(in_dir=img_dir)
    names_lst = list(pd.read_csv(names_path, header=None)[0])
    img_paths = glob.glob(img_dir + "/*.png")
    img_paths.sort(
        key=lambda v: [
            int(x) if x.isdigit() else x for x in re.findall(r"[^0-9]|[0-9]+", v)
        ]
    )
    for name, img_path in zip(names_lst, img_paths):
        lookup[name] = {}
        lookup[name]["img_path"] = img_path
    return lookup


def get_bp_config_codes() -> dict:
    """
    Helper to match SimBA project_config.ini [create ensemble settings][pose_estimation_body_parts] to string names.

    :return dict
    """

    return {
        "1 animal; 4 body-parts": "4",
        "1 animal; 7 body-parts": "7",
        "1 animal; 8 body-parts": "8",
        "1 animal; 9 body-parts": "9",
        "2 animals; 8 body-parts": "8",
        "2 animals; 14 body-parts": "14",
        "2 animals; 16 body-parts": "16",
        "MARS": Methods.USER_DEFINED.value,
        "Multi-animals; 4 body-parts": "8",
        "Multi-animals; 7 body-parts": "14",
        "Multi-animals; 8 body-parts": "16",
        "3D tracking": "3D_user_defined",
    }


def get_bp_config_code_class_pairs() -> dict:
    """
    Helper to match SimBA project_config.ini [create ensemble settings][pose_estimation_body_parts] to feature extraction module class.

    :return dict: Dictionary with [create ensemble settings][pose_estimation_body_parts] entry as keys and feature extraction classes as keys.
    """

    from simba.feature_extractors.feature_extractor_4bp import \
        ExtractFeaturesFrom4bps
    from simba.feature_extractors.feature_extractor_7bp import \
        ExtractFeaturesFrom7bps
    from simba.feature_extractors.feature_extractor_8bp import \
        ExtractFeaturesFrom8bps
    from simba.feature_extractors.feature_extractor_8bps_2_animals import \
        ExtractFeaturesFrom8bps2Animals
    from simba.feature_extractors.feature_extractor_9bp import \
        ExtractFeaturesFrom9bps
    from simba.feature_extractors.feature_extractor_14bp import \
        ExtractFeaturesFrom14bps
    from simba.feature_extractors.feature_extractor_16bp import \
        ExtractFeaturesFrom16bps
    from simba.feature_extractors.feature_extractor_user_defined import \
        UserDefinedFeatureExtractor

    return {
        "16": ExtractFeaturesFrom16bps,
        "14": ExtractFeaturesFrom14bps,
        "9": ExtractFeaturesFrom9bps,
        "8": {1: ExtractFeaturesFrom8bps, 2: ExtractFeaturesFrom8bps2Animals},
        "7": ExtractFeaturesFrom7bps,
        "4": ExtractFeaturesFrom4bps,
        "user_defined": UserDefinedFeatureExtractor,
    }


def get_icons_paths() -> dict:
    """
    Helper to get dictionary with icons and their paths.

    :return dict: Dictionary with icon names as keys and icon paths as values.
    """

    simba_dir = os.path.dirname(simba.__file__)
    icons_dir = os.path.join(simba_dir, Paths.ICON_ASSETS.value)
    icon_paths = glob.glob(icons_dir + "/*.png")
    icons = {}
    for icon_path in icon_paths:
        _, icon_name, _ = get_fn_ext(icon_path)
        icons[icon_name] = {}
        icons[icon_name]["icon_path"] = icon_path
    return icons


def get_third_party_appender_file_formats() -> dict:
    """
    Helper to get dictionary the file-types of third-party annotation tools.

    :return dict: Dictionary with tool as keys and file types as values.
    """

    return {
        "BORIS": "csv",
        "ETHOVISION": "xlsx",
        "OBSERVER": "xlsx",
        "SOLOMON": "csv",
        "DEEPETHOGRAM": "csv",
        "BENTO": "annot",
    }


def get_emojis() -> dict:
    """
    Helper to get dictionary of emojis with names as keys and emojis as values.
    """
    python_version = str(f"{sys.version_info.major}.{sys.version_info.minor}")
    if python_version == "3.6":
        return {
            "thank_you": "".join(
                chr(x) for x in struct.unpack(">2H", "\U0001f64f".encode("utf-16be"))
            ),
            "relaxed": "".join(
                chr(x) for x in struct.unpack(">2H", "\U0001F600".encode("utf-16be"))
            ),
            "error": "".join(
                chr(x) for x in struct.unpack(">2H", "\U0001F6A8".encode("utf-16be"))
            ),
            "complete": "".join(
                chr(x) for x in struct.unpack(">2H", "\U0001F680".encode("utf-16be"))
            ),
            "warning": "".join(
                chr(x) for x in struct.unpack(">2H", "\u2757\uFE0F".encode("utf-16be"))
            ),
            "trash": "".join(
                chr(x) for x in struct.unpack(">2H", "\U0001F5D1".encode("utf-16be"))
            ),
        }

    elif python_version == "3.10" or python_version == "3.9":
        return {
            "thank_you": "\U0001f64f".encode("utf-8", "replace").decode(),
            "relaxed": "\U0001F600".encode("utf-8", "replace").decode(),
            "warning": "\u2757\uFE0F".encode("utf-8", "replace").decode(),
            "error": "\U0001F6A8".encode("utf-8", "replace").decode(),
            "complete": "\U0001F680".encode("utf-8", "replace").decode(),
            "trash": "\U0001F5D1".encode("utf-8", "replace").decode(),
        }

    elif python_version == "3.7":
        return {
            "thank_you": "\U0001f64f".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
            "relaxed": "\U0001F600".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
            "error": "\U0001F6A8".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
            "complete": "\U0001F680".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
            "warning": "\u2757\uFE0F".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
            "trash": "\U0001F5D1F".encode("utf16", errors="surrogatepass").decode(
                "utf16"
            ),
        }

    else:
        return {
            "thank_you": "\U0001f64f",
            "relaxed": "\U0001F600",
            "error": "\U0001F6A8",
            "complete": "\U0001F680",
            "trash": "\U0001F5D1",
        }


def get_meta_data_file_headers() -> List[str]:
    """
    Get List of headers for SimBA classifier metadata output.

    :return List[str]:
    """

    return [
        "Classifier_name",
        "RF_criterion",
        "RF_max_features",
        "RF_min_sample_leaf",
        "RF_n_estimators",
        "compute_feature_permutation_importance",
        "generate_classification_report",
        "generate_example_decision_tree",
        "generate_features_importance_bar_graph",
        "generate_features_importance_log",
        "generate_precision_recall_curves",
        "generate_rf_model_meta_data_file",
        "generate_sklearn_learning_curves",
        "learning_curve_data_splits",
        "learning_curve_k_splits",
        "n_feature_importance_bars",
        "over_sample_ratio",
        "over_sample_setting",
        "train_test_size",
        "train_test_split_type",
        "under_sample_ratio",
        "under_sample_setting",
        "class_weight",
    ]


def get_cmaps() -> List[str]:
    """
    Get list of named matplotlib color palettes.
    """
    return [
        "spring",
        "summer",
        "autumn",
        "cool",
        "Wistia",
        "Pastel1",
        "Set1",
        "winter",
        "afmhot",
        "gist_heat",
        "copper",
    ]


def get_color_dict() -> dict:
    """
    Get dict of color names as keys and RGB tuples as values
    """

    return {
        "Grey": (220, 200, 200),
        "Red": (0, 0, 255),
        "Dark-red": (0, 0, 139),
        "Maroon": (0, 0, 128),
        "Orange": (0, 165, 255),
        "Dark-orange": (0, 140, 255),
        "Coral": (80, 127, 255),
        "Chocolate": (30, 105, 210),
        "Yellow": (0, 255, 255),
        "Green": (0, 128, 0),
        "Dark-grey": (105, 105, 105),
        "Light-grey": (192, 192, 192),
        "Pink": (178, 102, 255),
        "Lime": (204, 255, 229),
        "Purple": (255, 51, 153),
        "Cyan": (255, 255, 102),
        "White": (255, 255, 255),
        "Black": (0, 0, 0),
        "Darkgoldenrod": (184, 134, 11),
        "Olive": (109, 113, 46),
        "Seagreen": (46, 139, 87),
        "Dodgerblue": (30, 144, 255),
        "Springgreen": (0, 255, 127),
        "Firebrick": (178, 34, 34),
        "Indigo": (63, 15, 183),
    }


def get_named_colors() -> List[str]:
    """
    Get list of named matplotlib colors.
    """
    return [
        "red",
        "pink",
        "lime",
        "gold",
        "coral",
        "lavender",
        "sienna",
        "tomato",
        "grey",
        "azure",
        "crimson",
        "aqua",
        "plum",
        "blue",
        "teal",
        "maroon",
        "green",
        "black",
        "deeppink",
        "darkgoldenrod",
        "purple",
        "olive",
        "seagreen",
        "dodgerblue",
        "springgreen",
        "firebrick",
        "indigo" "white",
    ]


def create_color_palettes(no_animals: int, map_size: int) -> List[List[int]]:
    """
    Create list of lists of bgr colors, one for each animal. Each list is pulled from a different palette
    matplotlib color map.

    :param int no_animals: Number of different palette lists
    :param int map_size: Number of colors in each created palette.
    :return List[List[int]]:  BGR colors

    :example:
    >>> create_color_palettes(no_animals=2, map_size=2)
    >>> [[[255.0, 0.0, 255.0], [0.0, 255.0, 255.0]], [[102.0, 127.5, 0.0], [102.0, 255.0, 255.0]]]
    """
    colorListofList = []
    cmaps = [
        "spring",
        "summer",
        "autumn",
        "cool",
        "Wistia",
        "Pastel1",
        "Set1",
        "winter",
        "afmhot",
        "gist_heat",
        "copper",
    ]
    for colormap in range(no_animals):
        currColorMap = cm.get_cmap(cmaps[colormap], map_size)
        currColorList = []
        for i in range(currColorMap.N):
            rgb = list((currColorMap(i)[:3]))
            rgb = [i * 255 for i in rgb]
            rgb.reverse()
            currColorList.append(rgb)
        colorListofList.append(currColorList)
    return colorListofList


def cardinality_to_integer_lookup():
    return {"N": 0, "NE": 1, "E": 2, "SE": 3, "S": 4, "SW": 5, "W": 6, "NW": 7}


def integer_to_cardinality_lookup():
    return {0: "N", 1: "NE", 2: "E", 3: "SE", 4: "S", 5: "SW", 6: "W", 7: "NW"}


def percent_to_crf_lookup():
    return {
        "10": 37,
        "20": 34,
        "30": 31,
        "40": 28,
        "50": 25,
        "60": 22,
        "70": 19,
        "80": 16,
        "90": 13,
        "100": 10,
    }


def video_quality_to_preset_lookup():
    return {"Low": "fast", "Medium": "medium", "High": "slow"}


def get_log_config():
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "format": "%(asctime)s|%(name)s||%(message)s",
                "datefmt": "%Y-%m-%dT%H:%M:%SZ",
                # "class": "pythonjsonlogger.jsonlogger.JsonFormatter",
            }
        },
        "handlers": {
            "file_handler": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "default",
                "mode": "a",
                "backupCount": 5,
                "maxBytes": 5000000,
            }
        },
        "loggers": {"": {"level": "INFO", "handlers": ["file_handler"]}},
    }


#
# def rao_spacing_critical_values():
#     {4.0: 186.45,
#      5.0: 183.44,
#      6.0: 180.65,
#      7.0: 177.83,
#      8.0: 175.68,
#      9.0: 173.68,
#      10.0: 171.98,
#      11.0: 170.45,
#      12.0: 169.09,
#      13.0: 167.87,
#      14.0: 166.76,
#      15.0: 165.75,
#      16.0: 164.83,
#      17.0: 163.98,
#      18.0: 163.2,
#      19.0: 162.47,
#      20.0: 161.79,
#      21.0: 161.16,
#      22.0: 160.56,
#      23.0: 160.01,
#      24.0: 159.48,
#      25.0: 158.99,
#      26.0: 158.52,
#      27.0: 158.07,
#      28.0: 157.65,
#      29.0: 157.25,
#      30.0: 156.87,
#      35.0: 155.19,
#      40.0: 153.82,
#      45.0: 152.68,
#      50.0: 151.7,
#      75.0: 148.34,
#      100.0: 146.29,
#      150.0: 143.83,
#      200.0: 142.35,
#      300.0: 140.57,
#      400.0: 139.5,
#      500.0: 138.77,
#      600.0: 138.23,
#      700.0: 137.8,
#      800.0: 137.46,
#      900.0: 137.18,
#      1000.0: 136.94}



from tkinter import *

from simba.data_processors.boolean_conditional_calculator import \
    BooleanConditionalCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, Label, LabelFrame)
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.errors import CountError, DuplicationError
from simba.utils.read_write import read_df


class BooleanConditionalSlicerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(
            self, title="CONDITIONAL BOOLEAN AGGREGATE STATISTICS", size=(600, 400)
        )
        self.rule_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="CONDITIONAL RULES #",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.AGGREGATE_BOOL_STATS.value,
        )
        self.rule_cnt_dropdown = DropDownMenu(
            self.rule_cnt_frm,
            "# RULES:",
            list(range(2, 21)),
            "25",
            com=self.create_rules_frames,
        )
        self.rule_cnt_dropdown.setChoices(2)
        self.rule_cnt_frm.grid(row=0, column=0, sticky="NW")
        self.rule_cnt_dropdown.grid(row=0, column=0, sticky="NW")

        self.create_run_frm(run_function=self.run)
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg=f"No data found in {self.features_dir}",
        )
        data_df = read_df(
            file_path=self.feature_file_paths[0], file_type=self.file_type
        )
        self.bool_cols = data_df.columns[data_df.apply(self._is_bool)]
        if len(self.bool_cols) < 2:
            raise CountError(
                msg=f"The data file {self.feature_file_paths[0]} contains less than 2 boolean columns",
                source=self.__class__.__name__,
            )
        self.create_rules_frames(rules_cnt=2)
        self.main_frm.mainloop()

    @staticmethod
    def _is_bool(column):
        unique_values = set(column)
        return unique_values.issubset({0, 1})

    def create_rules_frames(self, rules_cnt: int):
        if hasattr(self, "rule_definitions_frame"):
            self.rule_definitions_frame.destroy()
        self.rule_definitions_frame = LabelFrame(
            self.main_frm,
            text="CONDITIONAL RULES",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.rule_definitions_frame.grid(row=1, column=0, sticky="NW")
        Label(self.rule_definitions_frame, text="RULE #").grid(
            row=0, column=0, sticky=NW
        )
        Label(self.rule_definitions_frame, text="BEHAVIOR").grid(
            row=0, column=1, sticky=NW, padx=5
        )
        Label(self.rule_definitions_frame, text="STATUS").grid(
            row=0, column=2, sticky=NW, padx=5
        )

        self.rules = {}
        for rule_cnt in range(1, rules_cnt + 1):
            self.rules[rule_cnt] = {}
            Label(
                self.rule_definitions_frame,
                text=str(rule_cnt),
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
            ).grid(row=rule_cnt, column=0, sticky=NW)
            self.rules[rule_cnt]["behavior_drpdwn"] = DropDownMenu(
                self.rule_definitions_frame,
                "",
                self.bool_cols,
                "1",
            )
            self.rules[rule_cnt]["behavior_drpdwn"].setChoices(
                self.bool_cols[rule_cnt - 1]
            )
            self.rules[rule_cnt]["behavior_drpdwn"].grid(
                row=rule_cnt, column=1, sticky=NW
            )
            self.rules[rule_cnt]["status_drpdwn"] = DropDownMenu(
                self.rule_definitions_frame, "", Options.BOOL_STR_OPTIONS.value, "1"
            )
            self.rules[rule_cnt]["status_drpdwn"].setChoices(
                Options.BOOL_STR_OPTIONS.value[0]
            )
            self.rules[rule_cnt]["status_drpdwn"].grid(
                row=rule_cnt, column=2, sticky=NW
            )

    def run(self):
        unique_rule_behaviors = []
        selections = {}
        for rule_id, rule_data in self.rules.items():
            unique_rule_behaviors.append(rule_data["behavior_drpdwn"].getChoices())
            selections[rule_data["behavior_drpdwn"].getChoices()] = rule_data[
                "status_drpdwn"
            ].getChoices()
        duplicates = list(
            set(
                [x for x in unique_rule_behaviors if unique_rule_behaviors.count(x) > 1]
            )
        )
        if len(duplicates) > 0:
            raise DuplicationError(
                msg=f"Each row should be a unique behavior. However, behaviors {duplicates} are selected in more than 1 rows."
            )
        boolean_calculator = BooleanConditionalCalculator(
            config_path=self.config_path, rules=selections
        )
        boolean_calculator.run()
        boolean_calculator.save()


# roi_featurizer = BooleanConditionalSlicerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini')
# roi_featurizer = BooleanConditionalSlicerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import functools
import multiprocessing
import os
import shutil
import subprocess

try:
    from typing import List
except:
    from typing_extensions import List

from simba.utils.checks import check_file_exist_and_readable, check_int
from simba.utils.read_write import get_fn_ext


class VideoProcessingMixin(object):
    """
    Methods for videos processing
    """

    def __init__(self):
        pass

    @staticmethod
    def _chunk_video_helper(chunk_range, video_path, chunk_ranges, temp_dir):
        start_time, end_time = chunk_range
        chunk_index = chunk_ranges.index(chunk_range)
        output_file = os.path.join(temp_dir, f"{chunk_index}.mp4")
        command = 'ffmpeg -i "{}" -ss {} -to {} -c copy "{}" -y'.format(
            video_path, start_time, end_time, output_file
        )
        subprocess.call(command, shell=True)
        return output_file

    @staticmethod
    def _ffmpeg_cmd_multiprocessor(command: str):
        print(command)
        subprocess.call(command, shell=True)

    def split_video_into_n_cunks(self, video_path: str, n: int):

        dir, video_name, _ = get_fn_ext(filepath=video_path)
        temp_dir = os.path.join(dir, video_name + "_temp")
        if os.path.isdir(temp_dir):
            shutil.rmtree(temp_dir)
        if not os.path.isdir(temp_dir):
            os.makedirs(temp_dir)
        os.makedirs(dir, exist_ok=True)
        duration = float(
            subprocess.check_output(
                'ffprobe -v error -show_entries format=duration -of default=noprint_wrappers=1:nokey=1 "{}" -hide_banner -loglevel error'.format(
                    video_path
                ),
                shell=True,
            )
        )
        chunk_duration = duration / n
        chunk_ranges = [
            (i * chunk_duration, (i + 1) * chunk_duration) for i in range(n)
        ]
        file_paths = []
        with multiprocessing.Pool(n, maxtasksperchild=10) as pool:
            constants = functools.partial(
                self._chunk_video_helper,
                video_path=video_path,
                chunk_ranges=chunk_ranges,
                temp_dir=temp_dir,
            )
            for cnt, result in enumerate(
                pool.imap(constants, chunk_ranges, chunksize=1)
            ):
                file_paths.append(result)
        pool.terminate()
        pool.join()
        return file_paths

    def create_ouput_paths(self, video_input_paths: List[os.PathLike]):

        temp_folder = os.path.join(os.path.dirname(video_input_paths[0]), "temp")
        if os.path.isdir(temp_folder):
            shutil.rmtree(temp_folder)
        if not os.path.isdir(temp_folder):
            os.makedirs(temp_folder)
        output_paths = [
            os.path.join(temp_folder, os.path.basename(x)) for x in video_input_paths
        ]
        return output_paths


# video_processor = VideoProcessingMixin()
#
# video_processor.change_single_video_fps(video_path='/Users/simon/Desktop/Example_1_frame_no.mp4', core_cnt=5)
#
#



import os.path, glob

import pytest
from simba.third_party_label_appenders.BORIS_appender import BorisAppender
from simba.third_party_label_appenders.solomon_importer import SolomonImporter
from simba.third_party_label_appenders.ethovision_import import ImportEthovision
from simba.third_party_label_appenders.observer_importer import NoldusObserverImporter
from simba.third_party_label_appenders.BENTO_appender import BentoAppender
from simba.third_party_label_appenders.deepethogram_importer import DeepEthogramImporter


@pytest.mark.parametrize("config_path, boris_path", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                      'tests/data/test_projects/two_c57/boris_annotations')])
def test_boris_import_use_case(config_path, boris_path):
    boris_appender = BorisAppender(config_path=config_path,
                                   data_dir=boris_path)
    boris_appender.create_boris_master_file()
    boris_appender.run()
    #assert len(boris_appender.out_df) == 1738
    #for f in glob.glob(boris_appender.targets_folder + '/*.csv'): os.remove(f)

@pytest.mark.parametrize("config_path, solomon_path", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                        'tests/data/test_projects/two_c57/solomon_annotations')])
def test_solomon_import_use_case(config_path, solomon_path):
    solomon_appender = SolomonImporter(config_path=config_path,
                                       data_dir=solomon_path)
    solomon_appender.run()

@pytest.mark.parametrize("config_path, data_dir", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                    'tests/data/test_projects/two_c57/ethovision_annotations')])
def test_ethovision_import_use_case(config_path, data_dir):
    ethovision_importer = ImportEthovision(config_path=config_path, data_dir=data_dir)
    ethovision_importer.run()
    for f in glob.glob(ethovision_importer.targets_folder + '/*.csv'): os.remove(f)

@pytest.mark.parametrize("config_path, data_dir", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                    'tests/data/test_projects/two_c57/observer_annotations')])
def test_observer_import_use_case(config_path, data_dir):
    observer_importer = NoldusObserverImporter(config_path=config_path, data_dir=data_dir)
    observer_importer.run()
    for f in glob.glob(observer_importer.targets_folder + '/*.csv'): os.remove(f)

@pytest.mark.parametrize("config_path, data_dir", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                    'tests/data/test_projects/two_c57/bento_annotations')])
def test_bento_import_use_case(config_path, data_dir):
    bento_appender = BentoAppender(config_path=config_path, data_dir=data_dir)
    bento_appender.run()
    for f in glob.glob(bento_appender.targets_folder + '/*.csv'): os.remove(f)

@pytest.mark.parametrize("config_path, data_dir", [('tests/data/test_projects/two_c57/project_folder/project_config.ini',
                                                    'tests/data/test_projects/two_c57/deepethogram_annotations')])
def test_deepethogram_import_use_case(config_path, data_dir):
    deepethogram_importer = DeepEthogramImporter(config_path=config_path, data_dir=data_dir)
    deepethogram_importer.run()
    for f in glob.glob(deepethogram_importer.targets_folder + '/*.csv'): os.remove(f)



__author__ = "Simon Nilsson"

import glob
import os
import warnings
from datetime import datetime

import pandas as pd
from tables import NaturalNameWarning

from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.errors import (BodypartColumnNotFoundError,
                                InvalidFileTypeError, NotDirectoryError)
from simba.utils.printing import SimbaTimer, stdout_success

warnings.filterwarnings("ignore", category=NaturalNameWarning)


class KeypointRemover(object):
    """
    Remove pose-estimated keypoints from data in CSV or H5 format.

    :param str data_folder: Path to directory containing pose-estiation CSV or H5 data
    :param str file_format: File type of pose-estimation data.
    :param str pose_tool: Tool used to perform pose-estimation. E.g., `DLC` or `maDLC`

    .. note:
    `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#remove-body-parts-from-tracking-data>`__.

    Examples
    ----------
    >>> keypoint_remover = KeypointRemover(data_folder="MyDataFolder", pose_tool='maDLC', file_format='h5')
    >>> keypoint_remover.run(bp_to_remove_list=['Nose_1, Nose_2'])
    """

    def __init__(self, data_folder: str, pose_tool: str, file_format: str):
        if not os.path.isdir(data_folder):
            raise NotDirectoryError(
                msg="{} is not a valid directory.".format(str(data_folder))
            )
        self.files_found = glob.glob(data_folder + "/*." + file_format)
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: Zero files found of type {} in the {} directory".format(
                file_format, data_folder
            ),
        )
        self.datetime = str(datetime.now().strftime("%Y%m%d%H%M%S"))
        self.pose_tool, self.data_folder = pose_tool, data_folder
        self.file_format = file_format
        if file_format == "h5":
            first_df = pd.read_hdf(self.files_found[0])
        else:
            first_df = pd.read_csv(self.files_found[0], header=[0, 1, 2])
        header_list = list(first_df.columns)[1:]
        self.body_part_names, self.animal_names = [], []

        if pose_tool == "DLC":
            for header_entry in header_list:
                if header_entry[1] not in self.body_part_names:
                    self.body_part_names.append(header_entry[1])

        else:
            for header_entry in header_list:
                if header_entry[1] not in self.body_part_names:
                    self.animal_names.append(header_entry[1])
                    self.body_part_names.append(header_entry[2])

        self.body_part_names, self.animal_names = list(set(self.body_part_names)), list(
            set(self.animal_names)
        )

    def run(self, animal_names: list, bp_to_remove_list: list):
        self.timer = SimbaTimer()
        self.timer.start_timer()
        save_directory = os.path.join(
            self.data_folder, "Reorganized_bp_{}".format(self.datetime)
        )
        if not os.path.exists(save_directory):
            os.makedirs(save_directory)
        print(
            "Saving {} new pose-estimation files in {} directory...".format(
                str(len(self.files_found)), save_directory
            )
        )
        if (self.pose_tool == "DLC") or (self.pose_tool == "maDLC"):
            for file_cnt, file_path in enumerate(self.files_found):
                save_path = os.path.join(save_directory, os.path.basename(file_path))
                if self.file_format == "csv":
                    self.df = pd.read_csv(file_path, header=[0, 1, 2], index_col=0)
                    for body_part in bp_to_remove_list:
                        if body_part not in self.df.columns._levels[1]:
                            raise BodypartColumnNotFoundError(
                                msg=f"{body_part} key point is not present in file {file_path}"
                            )
                        self.df = self.df.drop(body_part, axis=1, level=1)
                    self.df.to_csv(save_path)
                if self.file_format == "h5":
                    self.df = pd.read_hdf(file_path)
                    try:
                        first_header_value = self.df.columns._levels[0].values[0]
                    except:
                        raise InvalidFileTypeError(
                            msg=f"{file_path} is not a valid maDLC pose-estimation file"
                        )
                    for body_part, animal_name in zip(bp_to_remove_list, animal_names):
                        for cord in ["x", "y", "likelihood"]:
                            try:
                                self.df = self.df.drop(
                                    (first_header_value, animal_name, body_part, cord),
                                    axis=1,
                                )
                            except:
                                raise BodypartColumnNotFoundError(
                                    msg=f"Could not find body part {body_part} in {file_path}"
                                )
                    self.df.to_hdf(
                        save_path, key="re-organized", format="table", mode="w"
                    )
                print(
                    "Saved {}, Video {}/{}.".format(
                        os.path.basename(file_path),
                        str(file_cnt + 1),
                        str(len(self.files_found)),
                    )
                )
            self.timer.stop_timer()
            stdout_success(
                msg=f"{str(len(self.files_found))} new data with {str(len(bp_to_remove_list))} body-parts removed saved in {save_directory} directory",
                elapsed_time=self.timer.elapsed_time_str,
            )



import pytest
import numpy as np
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.roi_tools.ROI_clf_calculator import ROIClfCalculator
from simba.roi_tools.ROI_directing_analyzer import DirectingROIAnalyzer
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.roi_tools.ROI_movement_analyzer import ROIMovementAnalyzer
from simba.roi_tools.ROI_size_calculations import rectangle_size_calc, circle_size_calc, polygon_size_calc
from simba.roi_tools.ROI_time_bin_calculator import ROITimebinCalculator


@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini'])
def config_path_args(request):
    return request

@pytest.fixture(params=['tests/data/test_projects/two_c57/project_folder/project_config.ini',
                        'tests/data/test_projects/mouse_open_field/project_folder/project_config.ini',
                        'tests/data/test_projects/zebrafish/project_folder/project_config.ini'])
def all_config_path_args(request):
    return request


@pytest.mark.parametrize("calculate_distances", [True, False])
def test_roi_analyser(all_config_path_args, calculate_distances):
    roi_analyzer = ROIAnalyzer(ini_path=all_config_path_args.param,
                               data_path="outlier_corrected_movement_location",
                               calculate_distances=calculate_distances)
    roi_analyzer.run()
    roi_analyzer.save()

def test_roi_clf_calculator(config_path_args):
    clf_ROI_analyzer = ROIClfCalculator(config_ini=config_path_args.param)
    clf_ROI_analyzer.run(behavior_list=['Attack', 'Sniffing'], ROI_dict_lists={'Rectangle': ['Rectangle_1']}, body_part_list=['Nose_1'], measurements=['Total time by ROI (s)', 'Started bouts by ROI (count)', 'Ended bouts by ROI (count)'])

# @pytest.mark.parametrize("config_path, settings", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4})])
# def test_roi_directing_analyzer(config_path, settings):
#     directing_analyzer = DirectingROIAnalyzer(config_path=config_path, settings=settings, )
#     directing_analyzer.run()

# @pytest.mark.parametrize("config_path, settings", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', {'body_parts': {'Simon': 'Ear_left_1', 'JJ': 'Ear_left_2'}, 'threshold': 0.4})])
# def test_roi_feature_creator(config_path, settings):
#     roi_featurizer = ROIFeatureCreator(config_path=config_path, settings=settings)
#     roi_featurizer.run()

def test_roi_movement_analyzer(all_config_path_args):
    _ = ROIMovementAnalyzer(config_path=all_config_path_args.param)

@pytest.mark.parametrize("rectangle_dict, px_mm, expected_area", [({'height': 500, 'width': 500}, 10, 25), ({'height': 500, 'width': 500}, 5, 100)])
def test_rectangle_size_calc(rectangle_dict, px_mm, expected_area):
    results = rectangle_size_calc(rectangle_dict=rectangle_dict, px_mm=px_mm)
    assert int(results['area_cm']) == int(expected_area)

@pytest.mark.parametrize("circle_dict, px_mm, expected_area", [({'radius': 100}, 5, 12.57)])
def test_circle_size_calc(circle_dict, px_mm, expected_area):
    results = circle_size_calc(circle_dict=circle_dict, px_mm=px_mm)
    assert results['area_cm'] == expected_area

@pytest.mark.parametrize("polygon_dict, px_mm, expected_area", [({'vertices': np.array([[0, 2], [200, 98], [100, 876], [10, 702]])}, 5, 45.29)])
def test_polygon_size_calc(polygon_dict, px_mm, expected_area):
    results = polygon_size_calc(polygon_dict=polygon_dict, px_mm=px_mm)
    assert results['area_cm'] == expected_area


@pytest.mark.parametrize("bin_length, threshold", [(10, 0.00), (1, 0.50)])
def test_roi_timbin_calculator(config_path_args, bin_length, threshold):
    timebin_calculator = ROITimebinCalculator(config_path=config_path_args.param, bin_length=bin_length, body_parts=['Nose_1'], threshold=threshold)
    timebin_calculator.run()
    timebin_calculator.save()



import glob
import os
import shutil
from copy import deepcopy
from typing import Union

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import IdenticalInputWarning

SUBORDINATES = "subordinates"
WINNER = "winner"
TIE_BREAKER = "tie_breaker"
THRESHOLD = "threshold"
THRESHOLD_DETERMINATOR = "clf_threshold"
HIGHEST_PROBABILITY = "highest_probability"
SKIP_FILES_WITH_IDENTICAL = "skip_files_with_identical"


class MutualExclusivityCorrector(ConfigReader):
    """
    Refactor classification results according to user-defined mutual exclusivity rules.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/mutual_exclusivity_heuristic_rules.md>`__.


    Examples
    ----------
    >>> rules = {1: {'rule_type': 'threshold_determinator','determinator': 'Attack', 'threshold': 0.5, 'subordinates': ['Sniffing']}, 2: {'rule_type': 'threshold_determinator', 'determinator': 'Attack', 'threshold': 0.0, 'subordinates': ['Sniffing', 'Rear']}}
    >>> exclusivity_corrector = MutualExclusivityCorrector(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', rules=rules)
    >>> exclusivity_corrector.run()
    >>> rules = {1: {'rule_type': 'highest_probability', 'subordinates': ['body', 'face'], 'winner': 'body', 'skip_files_with_identical': True}}
    >>> exclusivity_corrector = MutualExclusivityCorrector(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', rules=rules)
    >>> exclusivity_corrector.run()
    """

    def __init__(self, rules: dict, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path)
        self.rules, self.save_dir = rules, None
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg=f"The {self.machine_results_dir} directory is empty",
        )
        self.save_dir = os.path.join(
            self.machine_results_dir, f"Prior_to_mutual_exclusivity_{self.datetime}"
        )

    def run(self):
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            self.video_name = get_fn_ext(filepath=file_path)[1]
            video_timer = SimbaTimer(start=True)
            print(f"Analysing mutual exclusivity in {self.video_name}..")
            self.input_df = read_df(file_path=file_path, file_type=self.file_type)
            self.data_df = deepcopy(self.input_df)
            for rule_cnt, rule_data in self.rules.items():
                self.rule_cnt, self.rule_data = rule_cnt, rule_data
                if rule_data["rule_type"] == THRESHOLD_DETERMINATOR:
                    check_that_column_exist(
                        df=self.data_df,
                        column_name=rule_data[WINNER],
                        file_name=file_path,
                    )
                    check_that_column_exist(
                        df=self.data_df,
                        column_name=f"Probability_{rule_data[WINNER]}",
                        file_name=file_path,
                    )
                    for subordinate in rule_data[SUBORDINATES]:
                        check_that_column_exist(
                            df=self.data_df,
                            column_name=f"Probability_{subordinate}",
                            file_name=file_path,
                        )
                    self.threshold_determinator()

                if rule_data["rule_type"] == HIGHEST_PROBABILITY:
                    self.probability_cols = [
                        f"Probability_{x}" for x in rule_data[SUBORDINATES]
                    ]
                    for subordinate in rule_data[SUBORDINATES]:
                        check_that_column_exist(
                            df=self.data_df,
                            column_name=subordinate,
                            file_name=file_path,
                        )
                    self.highest_probability_determinator()

            if not os.path.isdir(self.save_dir):
                os.makedirs(self.save_dir)
            shutil.move(
                file_path, os.path.join(self.save_dir, os.path.basename(file_path))
            )
            write_df(df=self.data_df, file_type=self.file_type, save_path=file_path)
            video_timer.stop_timer()
            stdout_success(
                msg=f"Mutual exclusivity complete video {self.video_name}. ({file_cnt+1}/{len(self.machine_results_paths)})",
                elapsed_time=video_timer.elapsed_time_str,
            )

        self.timer.stop_timer()
        stdout_success(
            msg=f"Mutual exclusivity performed on {len(self.machine_results_paths)} file(s). Results are saved in the project_folder/csv/machine_results directory. Copies of files PRIOR to applying mutual exclusivity rules are saved in {self.save_dir}",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def threshold_determinator(self):
        sums_df = self.data_df[
            self.rule_data[SUBORDINATES] + [self.rule_data[WINNER]]
        ].sum(axis=1)
        rows = list(
            sums_df.loc[(sums_df == len(self.rule_data[SUBORDINATES]) + 1)].index
        )
        rows = [
            x
            for x in rows
            if self.data_df.loc[x, f"Probability_{self.rule_data[WINNER]}"]
            > self.rule_data[THRESHOLD]
        ]
        for subordinate in self.rule_data[SUBORDINATES]:
            self.data_df.loc[rows, subordinate] = 0

    def highest_probability_determinator(self):
        sums = self.data_df[self.rule_data[SUBORDINATES]].sum(axis=1)
        overlap_df = self.data_df[self.probability_cols][
            self.data_df.index.isin(
                list(sums.index[sums == len(self.rule_data[SUBORDINATES])])
            )
        ]
        if len(overlap_df) > 0:
            identical_rows = list(
                overlap_df.index[overlap_df.apply(lambda x: min(x) == max(x), 1)]
            )
            if len(identical_rows) > 0 and self.rule_data[SKIP_FILES_WITH_IDENTICAL]:
                IdenticalInputWarning(
                    msg=f'Skipping rule {self.rule_cnt} for video {self.video_name} in frames {identical_rows}. Identical rows found and "SKIP ON EQUAL" selected...'
                )
                overlap_df = overlap_df.drop(index=identical_rows)
            elif (len(identical_rows) > 0) and not self.rule_data[
                SKIP_FILES_WITH_IDENTICAL
            ]:
                overlap_df.loc[
                    identical_rows, f"Probability_{self.rule_data[TIE_BREAKER]}"
                ] = (overlap_df[f"Probability_{self.rule_data[TIE_BREAKER]}"] + 1)
            overlap_array = (
                overlap_df.values == overlap_df.values.max(axis=1)[:, None]
            ).astype(int)
            results = pd.DataFrame(
                overlap_array,
                columns=self.rule_data[SUBORDINATES],
                index=overlap_df.index,
            )
            self.data_df.update(results)


# rules = {1: {'rule_type': 'clf_threshold',
#              'winner': 'Attack',
#              'threshold': 0.0,
#              'subordinates': ['Sniffing', 'Attack']},
#          2: {'rule_type': 'clf_threshold',
#              'winner': 'Sniffing',
#              'threshold': 0.0,
#              'subordinates': ['Sniffing', 'Rear', 'Attack']}}

# rules = {1: {'rule_type': 'clf_threshold',
#              'winner': 'Attack',
#              'threshold': 0,
#              'subordinates': ['Sniffing']},
#          2: {'rule_type': 'clf_threshold',
#              'winner': 'Sniffing',
#              'threshold': 0.0,
#              'subordinates': ['Rear', 'Attack']}}
#
# rules = {1: {'rule_type': 'highest_probability',
#              'tie_breaker': 'Attack',
#              'subordinates': ['Sniffing', 'Attack'],
#              'skip_files_with_identical': None},
#          2: {'rule_type': 'highest_probability',
#              'tie_breaker': 'Rear',
#              'subordinates': ['Rear', 'Attack'],
#              'skip_files_with_identical': None}}

# rules = {1: {'rule_type': 'highest_probability', 'subordinates': ['Sniffing', 'Attack'], 'tie_breaker': 'Attack', 'skip_files_with_identical': False}}
# test = MutualExclusivityCorrector(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', rules=rules)
# test.run()



import os
from typing import Union

import cv2
import numpy as np

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int)
from simba.utils.enums import Formats
from simba.utils.lookups import integer_to_cardinality_lookup
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_df)


class CircularFeaturePlotter(ConfigReader, PlottingMixin, FeatureExtractionMixin):
    """
    Create visualization of base angular features overlay on video. E.g., use to confirm
    accurate cardinality and angle degree computation.

    .. image:: _static/img/circular_visualiation.gif
       :width: 600
       :align: center

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] data_path: Path to file containing angular features.
    :param dict settings: Dictionary containing visualization attributes.

    :example:
    >>> settings = {'center': {'Animal_1': 'SwimBladder'}, 'text_settings': False, "palette": 'bwr'}
    >>> circular_feature_plotter = CircularFeaturePlotter(config_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/project_config.ini', data_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/csv/circular_features/20200730_AB_7dpf_850nm_0002.csv', settings=settings)
    >>> circular_feature_plotter.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        settings: dict,
    ):
        PlottingMixin.__init__(self)
        ConfigReader.__init__(self, config_path=config_path, read_video_info=True)
        FeatureExtractionMixin.__init__(self)
        self.video_path = find_video_of_file(
            video_dir=self.video_dir, filename=get_fn_ext(filepath=data_path)[1]
        )
        _, self.video_name, _ = get_fn_ext(filepath=self.video_path)
        check_file_exist_and_readable(file_path=data_path)
        self.data_path, self.config_path, self.settings, self.text_settings = (
            data_path,
            config_path,
            settings,
            settings["text_settings"],
        )
        self.save_path = os.path.join(
            self.project_path,
            "frames",
            "output",
            "circular_features",
            get_fn_ext(filepath=data_path)[1] + ".mp4",
        )
        if not os.path.isdir(os.path.dirname(self.save_path)):
            os.makedirs(os.path.dirname(self.save_path))
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)

    def _approximate_size_of_animal_in_video(self):
        self.animal_sizes = {}
        for animal_name, animal_bps in self.animal_bp_dict.items():
            self.animal_sizes[animal_name] = {}
            animal_bp_cols = [
                item
                for pair in zip(animal_bps["X_bps"], animal_bps["Y_bps"])
                for item in pair
            ]
            animal_bp_data = (
                self.df[animal_bp_cols]
                .values.reshape(len(self.df), len(animal_bps["X_bps"]), 2)
                .astype(np.float32)
            )
            self.animal_sizes[animal_name]["area"] = np.nanmean(
                jitted_hull(points=animal_bp_data, target="area")
            ).astype(np.int32)
            self.animal_sizes[animal_name]["diameter"] = (
                np.sqrt(self.animal_sizes[animal_name]["area"] / np.pi).astype(np.int32)
                * 3
            )

    def __get_print_settings(self):
        if self.text_settings is False:
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.1,
            )
            self.max_dim = max(
                self.video_meta_data["width"], self.video_meta_data["height"]
            )
            self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
            self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
            self.spacing_scale = int(self.space_scale / (self.res_scale / self.max_dim))
            self.text_thickness = 3
        else:
            check_float(name="ERROR: TEXT SIZE", value=self.text_settings["font_size"])
            check_int(name="ERROR: SPACE SIZE", value=self.text_settings["space_size"])
            check_int(
                name="ERROR: TEXT THICKNESS", value=self.text_settings["text_thickness"]
            )
            check_int(
                name="ERROR: CIRCLE SIZE", value=self.text_settings["circle_size"]
            )
            self.font_size = float(self.text_settings["font_size"])
            self.spacing_scale = int(self.text_settings["space_size"])
            self.text_thickness = int(self.text_settings["text_thickness"])
            self.circle_scale = int(self.text_settings["circle_size"])

    def __calc_text_locs(self):
        add_spacer = 2
        self.loc_dict = {}
        for animal_cnt, animal_name in enumerate(self.multi_animal_id_list):
            self.loc_dict[animal_name] = {}
            self.loc_dict[animal_name]["degree_txt"] = f"{animal_name} angle:"
            self.loc_dict[animal_name]["cardinal_txt"] = f"{animal_name} cardinal:"
            self.loc_dict[animal_name]["degree_txt_loc"] = (
                5,
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scale * add_spacer
                ),
            )
            self.loc_dict[animal_name]["degree_data_loc"] = (
                200,
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scale * add_spacer
                ),
            )
            add_spacer += 1
            self.loc_dict[animal_name]["cardinal_txt_loc"] = (
                5,
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scale * add_spacer
                ),
            )
            self.loc_dict[animal_name]["cardinal_data_loc"] = (
                240,
                (
                    self.video_meta_data["height"]
                    - (self.video_meta_data["height"] + 10)
                    + self.spacing_scale * add_spacer
                ),
            )
            add_spacer += 1

    def _create_palette(self):
        c = self.create_single_color_lst(
            pallete_name=self.settings["palette"], increments=360
        )
        self.colors = {}
        for color_cnt, color in enumerate(c):
            self.colors[color_cnt] = color

    def run(self):
        self.df = read_df(file_path=self.data_path, file_type=self.file_type)
        self.video_meta_data, frm_cnt = (
            get_video_meta_data(video_path=self.video_path),
            0,
        )
        self.__get_print_settings()
        self.cap = cv2.VideoCapture(self.video_path)
        self.writer = cv2.VideoWriter(
            self.save_path,
            self.fourcc,
            self.video_meta_data["fps"],
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        self._approximate_size_of_animal_in_video()
        self._create_palette()
        self.__calc_text_locs()
        for animal_name, animal_bps in self.animal_bp_dict.items():
            self.df[f"Compass_cardinal_{animal_name}"] = self.df[
                f"Compass_cardinal_{animal_name}"
            ].map(integer_to_cardinality_lookup())
        while self.cap.isOpened():
            ret, self.frame = self.cap.read()
            if not ret:
                break
            frm_data = self.df.iloc[frm_cnt, :]
            for animal_name, animal_bps in self.animal_bp_dict.items():
                compass_center = tuple(
                    frm_data[
                        [
                            f'{self.settings["center"][animal_name]}_x',
                            f'{self.settings["center"][animal_name]}_y',
                        ]
                    ].values.astype(int)
                )
                animal_frm_angle, compass_cardinal = (
                    int(frm_data[f"Angle_degrees_{animal_name}"]),
                    frm_data[f"Compass_cardinal_{animal_name}"],
                )
                frm_compass_clr = self.colors[animal_frm_angle]
                cv2.circle(
                    self.frame,
                    compass_center,
                    self.animal_sizes[animal_name]["diameter"],
                    frm_compass_clr,
                    self.text_thickness,
                )
                cv2.putText(
                    self.frame,
                    self.loc_dict[animal_name]["degree_txt"],
                    self.loc_dict[animal_name]["degree_txt_loc"],
                    self.font,
                    self.font_size,
                    frm_compass_clr,
                    1,
                )
                cv2.putText(
                    self.frame,
                    str(animal_frm_angle),
                    self.loc_dict[animal_name]["degree_data_loc"],
                    self.font,
                    self.font_size,
                    frm_compass_clr,
                    1,
                )
                cv2.putText(
                    self.frame,
                    self.loc_dict[animal_name]["cardinal_txt"],
                    self.loc_dict[animal_name]["cardinal_txt_loc"],
                    self.font,
                    self.font_size,
                    frm_compass_clr,
                    1,
                )
                cv2.putText(
                    self.frame,
                    compass_cardinal,
                    self.loc_dict[animal_name]["cardinal_data_loc"],
                    self.font,
                    self.font_size,
                    frm_compass_clr,
                    1,
                )

            self.writer.write(self.frame)
            print(
                f'Image {frm_cnt+1}/{self.video_meta_data["frame_count"]} Video: {self.video_path}'
            )
            frm_cnt += 1
        self.timer.stop_timer()
        self.writer.release()
        stdout_success(
            msg=f"Video {self.save_path} complete!",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# settings = {'center': {'Animal_1': 'SwimBladder'},
#             'text_settings': False, "palette": 'bwr'}
#
# test = CircularFeaturePlotter(config_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/project_config.ini',
#                               data_path='/Users/simon/Desktop/envs/troubleshooting/circular_features_zebrafish/project_folder/csv/circular_features/20200730_AB_7dpf_850nm_0002.csv',
#                               settings=settings)
# test.run()



__author__ = "Simon Nilsson"

import glob
import os
from collections import OrderedDict
from datetime import datetime
from typing import Optional

import pandas as pd

from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.enums import Formats
from simba.utils.printing import stdout_success


class KeypointReorganizer(object):
    """
    Re-organizing the order of pose-estimated keypoints in directory containing
    CSV or H5 format files.

    :param str data_folder: Path to directory containing pose-estiation CSV or H5 data
    :param str pose_tool: Tool used to perform pose-estimation. E.g., 'DLC' or 'maDLC'.
    :param str file_format: File type of pose-estimation data.

    .. note::
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#re-organize-tracking-data>`__.

    Examples
    ----------
    >>> keypoint_reorganizer = KeypointReorganizer(data_folder="test_data/misc_test_files", pose_tool='maDLC', file_format='h5')
    >>> keypoint_reorganizer.run(animal_list=['UM', 'LM', 'LM', 'UM', 'LM', 'UM', 'LM', 'LM', 'UM', 'LM', 'UM', 'UM', 'UM', 'UM', 'LM', 'LM'], bp_lst=['Lateral_left', 'Nose', 'Tail_base', 'Lateral_right', 'Ear_right', 'Center', 'Nose', 'Ear_left', 'Ear_right', 'Center', 'Tail_end', 'Ear_left', 'Tail_base', 'Lateral_left', 'Tail_end', 'Lateral_right'])
    >>> keypoint_reorganizer = KeypointReorganizer(data_folder="test_data/misc_test_files", pose_tool='DLC', file_format='csv')
    >>> keypoint_reorganizer.run(bp_lst=['Ear_left_1', 'Ear_right_1', 'Nose_1', 'Center_1', 'Lateral_left_1', 'Lateral_right_1', 'Tail_base_1', 'Ear_left_2', 'Ear_right_2', 'Nose_2', 'Center_2', 'Lateral_left_2', 'Lateral_right_2', 'Tail_base_2'], animal_list=None)
    """

    def __init__(
        self, data_folder: str, pose_tool: str, file_format: Optional[str] = None
    ):
        check_if_dir_exists(in_dir=data_folder)
        self.data_folder, self.pose_tool = data_folder, pose_tool
        self.file_format = file_format
        self.files_found = glob.glob(self.data_folder + "/*." + file_format)
        check_if_filepath_list_is_empty(
            self.files_found,
            error_msg=f"SIMBA ERROR: {self.data_folder} directory contains no {file_format} files.",
        )
        self.datetime = str(datetime.now().strftime("%Y%m%d%H%M%S"))

        if (file_format == Formats.H5.value) and (pose_tool == "DLC" or "maDLC"):
            first_df = pd.read_hdf(self.files_found[0])
            self.header_list = list(first_df.columns)
            animal_bp_tuple_list = []
            for header_val in self.header_list:
                new_value = (header_val[1], header_val[2])
                if new_value not in animal_bp_tuple_list:
                    animal_bp_tuple_list.append(new_value)
            animal_bp_tuple_list = list(set(animal_bp_tuple_list))
            self.animal_list = [x[0] for x in animal_bp_tuple_list]
            self.bp_list = [x[1] for x in animal_bp_tuple_list]

        elif (file_format == Formats.CSV.value) and (pose_tool == "DLC" or "maDLC"):
            first_df = pd.read_csv(self.files_found[0])
            if first_df.scorer.loc[2] == "coords":
                scorer = list(first_df.columns)
                scorer.remove("scorer")
                individuals = list(first_df.loc[0, :])
                individuals.remove("individuals")
                bodyparts = list(first_df.loc[1, :])
                bodyparts.remove("bodyparts")
                coords = list(first_df.loc[2, :])
                coords.remove("coords")
                self.header_list = list(zip(scorer, individuals, bodyparts, coords))
                animallist = list(zip(individuals, bodyparts))
                animal_bp_tuple_list = list(OrderedDict.fromkeys(animallist))
                self.animal_list = [x[0] for x in animal_bp_tuple_list]
                self.bp_list = [x[1] for x in animal_bp_tuple_list]
            else:
                self.animal_list = None
                scorer = list(first_df.columns)
                scorer.remove("scorer")
                bodyparts = list(first_df.loc[0, :])
                bodyparts.remove("bodyparts")
                self.bp_list = []
                for bp in bodyparts:
                    if bp not in self.bp_list:
                        self.bp_list.append(bp)
                coords = list(first_df.loc[1, :])
                coords.remove("coords")
                self.header_list = list(zip(scorer, bodyparts, coords))

        elif (file_format == Formats.CSV.value) and (pose_tool == "SLEAP"):
            print("s")

    def run(self, bp_lst: list, animal_list: list = None):
        save_directory = os.path.join(
            self.data_folder, "Reorganized_bp_{}".format(self.datetime)
        )
        if not os.path.exists(save_directory):
            os.makedirs(save_directory)
        print(
            "Saving {} new pose-estimation files in {} directory...".format(
                str(len(self.files_found)), save_directory
            )
        )
        header_tuples = []
        if self.pose_tool == "maDLC":
            for animal_name, animal_bp in zip(animal_list, bp_lst):
                header_tuples.append(
                    (self.header_list[0][0], animal_name, animal_bp, "x")
                )
                header_tuples.append(
                    (self.header_list[0][0], animal_name, animal_bp, "y")
                )
                header_tuples.append(
                    (self.header_list[0][0], animal_name, animal_bp, "likelihood")
                )
            new_df_ordered_cols = pd.MultiIndex.from_tuples(
                header_tuples, names=["scorer", "individuals", "bodyparts", "coords"]
            )
        if self.pose_tool == "DLC":
            for animal_bp in bp_lst:
                header_tuples.append((self.header_list[0][0], animal_bp, "x"))
                header_tuples.append((self.header_list[0][0], animal_bp, "y"))
                header_tuples.append((self.header_list[0][0], animal_bp, "likelihood"))
            new_df_ordered_cols = pd.MultiIndex.from_tuples(
                header_tuples, names=["scorer", "bodyparts", "coords"]
            )

        for file_cnt, file_path in enumerate(self.files_found):
            df_save_path = os.path.join(save_directory, os.path.basename(file_path))
            if self.file_format == "h5":
                df = pd.read_hdf(file_path)
                df_reorganized = pd.DataFrame(df, columns=new_df_ordered_cols)
                df_reorganized.to_hdf(
                    df_save_path, key="re-organized", format="table", mode="w"
                )
            if self.file_format == "csv":
                df = pd.read_csv(file_path, header=[0, 1, 2])
                df_reorganized = pd.DataFrame(df, columns=new_df_ordered_cols)
                df_reorganized.to_csv(df_save_path)
            print(
                "Saved {}, Video {}/{}.".format(
                    os.path.basename(file_path),
                    str(file_cnt + 1),
                    str(len(self.files_found)),
                )
            )
        stdout_success(
            msg=f"{str(len(self.files_found))} new data files with reorganized body-parts saved in {save_directory} directory"
        )


# keypoint_reorganizer = KeypointReorganizer(data_folder="/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5", pose_tool='SLEAP', file_format='csv')
# keypoint_reorganizer = KeypointReorganizer(data_folder="/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5", pose_tool='SLEAP', file_format='SLEAP H5')
# keypoint_reorganizer = KeypointReorganizer(data_folder="/Users/simon/Desktop/troubleshooting/B1-MS_US/el_import", pose_tool='maDLC', file_format='h5')
# keypoint_reorganizer.run(animal_list=['UM', 'LM', 'LM', 'UM', 'LM', 'UM', 'LM', 'LM', 'UM', 'LM', 'UM', 'UM', 'UM', 'UM', 'LM', 'LM'], bp_lst=['Nose', 'Tail_base', 'Tail_base', 'Lateral_right', 'Ear_right', 'Center', 'Nose', 'Ear_left', 'Ear_right', 'Center', 'Tail_end', 'Ear_left', 'Tail_base', 'Lateral_left', 'Tail_end', 'Lateral_right'])

# keypoint_reorganizer = KeypointReorganizer(data_folder="//Users/simon/Desktop/simbapypi_dev/tests/test_data/misc_test_files", pose_tool='DLC', file_format='csv')
# keypoint_reorganizer.run(bp_lst=['Ear_left_1', 'Ear_right_1', 'Nose_1', 'Center_1', 'Lateral_left_1', 'Lateral_right_1', 'Tail_base_1', 'Ear_left_2', 'Ear_right_2', 'Nose_2', 'Center_2', 'Lateral_left_2', 'Lateral_right_2', 'Tail_base_2'], animal_list=None)



__author__ = "Simon Nilsson"

import glob
import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import shutil
from itertools import combinations

from simba.feature_extractors.perimeter_jit import jitted_hull
from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_filepath_list_is_empty,
                                check_if_headers_in_dfs_are_unique,
                                check_same_number_of_rows_in_dfs,
                                check_that_column_exist)
from simba.utils.enums import Formats, Options
from simba.utils.errors import CountError, DataHeaderError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class FeatureSubsetsCalculator(ConfigReader, FeatureExtractionMixin, TrainModelMixin):
    """
    Computes a subset of features from pose for non-ML downstream purposes.
    E.g., returns the size of animal convex hull in each frame.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str save_dir: directory where to store results.
    :parameter str feature_family: Feature subtype to calculate. E.g., "Two-point body-part distances (mm)".

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/feature_subsets.md>_`

    .. image:: _static/img/feature_subsets.png
       :width: 400
       :align: center

    Examples
    ----------
    >>> _ = FeatureSubsetsCalculator(config_path='project_folder/project_config.ini', feature_family='Frame-by-frame body-parts inside ROIs (Boolean)', save_dir='data').run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        save_dir: Union[str, os.PathLike, None],
        feature_families: List[str],
        include_file_checks: Optional[bool] = False,
        append_to_features_extracted: Optional[bool] = False,
        append_to_targets_inserted: Optional[bool] = False,
    ):
        FeatureExtractionMixin.__init__(self, config_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg=f"SIMBA ERROR: Zero data files found in {self.outlier_corrected_paths} directory",
        )
        self.feature_families, self.save_dir = feature_families, save_dir
        self.append_to_features, self.append_to_targets, self.checks_files = (
            append_to_features_extracted,
            append_to_targets_inserted,
            include_file_checks,
        )
        self.create_temp_folders()

    def create_temp_folders(self):
        self.general_temp_folder = os.path.join(
            self.project_path, "csv", f"temp_data_{self.datetime}"
        )
        if not os.path.isdir(self.general_temp_folder):
            os.makedirs(self.general_temp_folder)
        self.features_extracted_temp_path, self.targets_inserted_temp_path = None, None
        if self.append_to_features or self.append_to_targets:
            if self.append_to_features:
                self.features_extracted_temp_path = os.path.join(
                    self.features_dir, f"temp_data_{self.datetime}"
                )
                if not os.path.isdir(self.features_extracted_temp_path):
                    os.makedirs(self.features_extracted_temp_path)
            if self.append_to_targets:
                self.targets_inserted_temp_path = os.path.join(
                    self.targets_folder, f"temp_data_{self.datetime}"
                )
                if not os.path.isdir(self.targets_inserted_temp_path):
                    os.makedirs(self.targets_inserted_temp_path)

    def perform_clean_up(self, directories: List[str]):
        print("Performing clean up and deleting temporary directories...")
        self.remove_multiple_folders(folders=directories)

    def __get_bp_combinations(self):
        self.two_point_combs = np.array(list(combinations(self.project_bps, 2)))
        self.within_animal_three_point_combs = {}
        self.within_animal_four_point_combs = {}
        self.animal_bps = {}
        for animal, animal_data in self.animal_bp_dict.items():
            animal_bps = [x[:-2] for x in animal_data["X_bps"]]
            self.animal_bps[animal] = animal_bps
            self.within_animal_three_point_combs[animal] = np.array(
                list(combinations(animal_bps, 3))
            )
            self.within_animal_four_point_combs[animal] = np.array(
                list(combinations(animal_bps, 4))
            )

    def run(self):
        self.__get_bp_combinations()
        for file_cnt, file_path in enumerate(self.outlier_corrected_paths):
            self.current_file_path = file_path
            self.video_timer = SimbaTimer(start=True)
            self.results = pd.DataFrame()
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            _, self.pixel_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.df = read_df(file_path=file_path, file_type=self.file_type)
            for family_cnt, feature_family in enumerate(self.feature_families):
                print(
                    f"Analyzing {self.video_name} and {feature_family} (Video {file_cnt+1}/{len(self.outlier_corrected_paths)}, Family {family_cnt+1}/{len(self.feature_families)})..."
                )
                if feature_family == "Two-point body-part distances (mm)":
                    self.calc_distances()
                elif (
                    feature_family
                    == "Within-animal three-point body-part angles (degrees)"
                ):
                    self.calc_angles()
                elif (
                    feature_family
                    == "Within-animal three-point convex hull perimeters (mm)"
                ):
                    self.calc_three_point_hulls()
                elif (
                    feature_family
                    == "Within-animal four-point convex hull perimeters (mm)"
                ):
                    self.calc_four_point_hulls()
                elif feature_family == "Entire animal convex hull perimeters (mm)":
                    self.animal_convex_hulls()
                elif feature_family == "Entire animal convex hull area (mm2)":
                    self.calc_animal_convex_hulls_area()
                elif feature_family == "Frame-by-frame body-part movements (mm)":
                    self.calc_movements()
                elif (
                    feature_family
                    == "Frame-by-frame body-part distances to ROI centers (mm)"
                ):
                    self.calc_roi_center_distances()
                elif (
                    feature_family == "Frame-by-frame body-parts inside ROIs (Boolean)"
                ):
                    self.calc_inside_roi()
            self.__save_to_temp()
            self.video_timer.stop_timer()
            print(
                f"Additional features computed for {self.video_name} complete (elapsed time {self.video_timer.elapsed_time_str}s)..."
            )
        if self.append_to_features or self.append_to_targets:
            print("Appending data to existing files...")
            self.append_to_data()
        self.replace_files_in_folders()
        self.perform_clean_up(
            directories=[
                self.targets_inserted_temp_path,
                self.features_extracted_temp_path,
                self.general_temp_folder,
            ]
        )
        self.timer.stop_timer()
        stdout_success(
            msg="Feature sub-sets calculations complete!",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def calc_distances(self):
        for c in self.two_point_combs:
            col_names = list(sum([(x + "_x", y + "_y") for (x, y) in zip(c, c)], ()))

            self.results[f"Distance (mm) {c[0]}-{c[1]}"] = self.euclidean_distance(
                self.df[col_names[0]].values,
                self.df[col_names[2]].values,
                self.df[col_names[1]].values,
                self.df[col_names[3]].values,
                self.pixel_per_mm,
            )

    def calc_angles(self):
        for animal, points in self.within_animal_three_point_combs.items():
            for point in points:
                col_names = list(
                    sum([(x + "_x", y + "_y") for (x, y) in zip(point, point)], ())
                )
                self.results[f"Angle (degrees) {point[0]}-{point[1]}-{point[2]}"] = (
                    self.angle3pt_serialized(data=self.df[col_names].values)
                )

    def calc_three_point_hulls(self):
        for animal, points in self.within_animal_three_point_combs.items():
            for point in points:
                col_names = list(
                    sum([(x + "_x", y + "_y") for (x, y) in zip(point, point)], ())
                )
                three_point_arr = np.reshape(
                    self.df[col_names].values, (len(self.df / 2), -1, 2)
                ).astype(np.float32)
                self.results[
                    f"{animal} three-point convex hull perimeter (mm) {point[0]}-{point[1]}-{point[2]}"
                ] = (
                    jitted_hull(points=three_point_arr, target=Formats.PERIMETER.value)
                    / self.pixel_per_mm
                )

    def calc_four_point_hulls(self):
        for animal, points in self.within_animal_four_point_combs.items():
            for point in points:
                col_names = list(
                    sum([(x + "_x", y + "_y") for (x, y) in zip(point, point)], ())
                )
                four_point_arr = np.reshape(
                    self.df[col_names].values, (len(self.df / 2), -1, 2)
                ).astype(np.float32)
                self.results[
                    f"{animal} four-point convex perimeter (mm) {point[0]}-{point[1]}-{point[2]}-{point[3]}"
                ] = (
                    jitted_hull(points=four_point_arr, target=Formats.PERIMETER.value)
                    / self.pixel_per_mm
                )

    def animal_convex_hulls(self):
        for animal, point in self.animal_bps.items():
            col_names = list(
                sum([(x + "_x", y + "_y") for (x, y) in zip(point, point)], ())
            )
            animal_point_arr = np.reshape(
                self.df[col_names].values, (len(self.df / 2), -1, 2)
            ).astype(np.float32)
            self.results[f"{animal} convex hull perimeter (mm)"] = (
                jitted_hull(points=animal_point_arr, target=Formats.PERIMETER.value)
                / self.pixel_per_mm
            )

    def calc_movements(self):
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_x", file_name=self.current_file_path
                )
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_y", file_name=self.current_file_path
                )
                bp_df = self.df[[f"{bp}_x", f"{bp}_y"]]
                shift = bp_df.shift(periods=1).add_suffix("_shift")
                bp_df = pd.concat([bp_df, shift], axis=1, join="inner").reset_index(
                    drop=True
                )
                for c in shift.columns:
                    bp_df[c] = bp_df[c].fillna(bp_df[c[:-6]])
                self.results[f"{animal} movement {bp} (mm)"] = self.euclidean_distance(
                    bp_df[f"{bp}_x_shift"].values,
                    bp_df[f"{bp}_x"].values,
                    bp_df[f"{bp}_y_shift"].values,
                    bp_df[f"{bp}_y"].values,
                    self.pixel_per_mm,
                )

    def calc_animal_convex_hulls_area(self):
        for animal, point in self.animal_bps.items():
            col_names = list(
                sum([(x + "_x", y + "_y") for (x, y) in zip(point, point)], ())
            )
            animal_point_arr = np.reshape(
                self.df[col_names].values, (len(self.df / 2), -1, 2)
            ).astype(np.float32)
            self.results[f"{animal} convex hull area (mm2)"] = (
                jitted_hull(points=animal_point_arr, target=Formats.AREA.value)
                / self.pixel_per_mm
            )

    def calc_roi_center_distances(self):
        self.read_roi_data()
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_x", file_name=self.current_file_path
                )
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_y", file_name=self.current_file_path
                )
                bp_arr = self.df[[f"{bp}_x", f"{bp}_y"]].values.astype(float)
                for shape_type, shape_data in self.roi_dict.items():
                    for shape_name in shape_data["Name"].unique():
                        center_point = (
                            shape_data.loc[
                                (shape_data["Video"] == self.video_name)
                                & (shape_data["Name"] == shape_name)
                            ][["Center_X", "Center_Y"]]
                            .astype(float)
                            .values[0]
                        )
                        self.results[
                            f"{animal} {bp} to {shape_name} center distance (mm)"
                        ] = self.framewise_euclidean_distance_roi(
                            location_1=bp_arr,
                            location_2=center_point,
                            px_per_mm=self.pixel_per_mm,
                        ).astype(
                            "float32"
                        )

    def calc_inside_roi(self):
        self.read_roi_data()
        for animal, animal_bps in self.animal_bps.items():
            for bp in animal_bps:
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_x", file_name=self.current_file_path
                )
                check_that_column_exist(
                    df=self.df, column_name=f"{bp}_y", file_name=self.current_file_path
                )
                bp_arr = self.df[[f"{bp}_x", f"{bp}_y"]].values.astype(float)
                for shape_type, shape_info in self.roi_dict.items():
                    for shape_name in shape_info["Name"].unique():
                        if shape_type == "rectangles":
                            shape_data = shape_info.loc[
                                (shape_info["Video"] == self.video_name)
                                & (shape_info["Name"] == shape_name)
                            ][
                                [
                                    "topLeftX",
                                    "topLeftY",
                                    "Bottom_right_X",
                                    "Bottom_right_Y",
                                ]
                            ].values.reshape(
                                2, 2
                            )
                            self.results[
                                f"{animal} {bp} inside rectangle {shape_name} (Boolean)"
                            ] = self.framewise_inside_rectangle_roi(
                                bp_location=bp_arr, roi_coords=shape_data
                            )
                        if shape_type == "polygons":
                            shape_data = shape_info.loc[
                                (shape_info["Video"] == self.video_name)
                                & (shape_info["Name"] == shape_name)
                            ][["vertices"]].values[0][0]
                            self.results[
                                f"{animal} {bp} inside polygon {shape_name} (Boolean)"
                            ] = self.framewise_inside_polygon_roi(
                                bp_location=bp_arr, roi_coords=shape_data
                            )

    def __save_to_temp(self):
        save_path = os.path.join(
            self.general_temp_folder, self.video_name + f".{self.file_type}"
        )
        write_df(
            df=self.results.fillna(-1), file_type=self.file_type, save_path=save_path
        )

    def append_to_data(self):
        for file_path in glob.glob(self.general_temp_folder + f"/*.{self.file_type}"):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            new_df = read_df(file_path=file_path, file_type=self.file_type)
            if self.append_to_features:
                print(f"Appending new data to features extracted from {video_name}...")
                data_save_path = os.path.join(
                    self.features_extracted_temp_path, video_name + f".{self.file_type}"
                )
                features_path = os.path.join(
                    self.features_dir, f"{video_name}.{self.file_type}"
                )
                check_file_exist_and_readable(file_path=file_path)
                data_df = read_df(file_path=features_path, file_type=self.file_type)
                if not check_same_number_of_rows_in_dfs(dfs=[data_df, new_df]):
                    self.remove_multiple_folders(
                        folders=[
                            self.targets_inserted_temp_path,
                            self.features_extracted_temp_path,
                            self.general_temp_folder,
                        ]
                    )
                    raise CountError(
                        f"Cannot append feature subsets to features file of {video_name}. The files contain an unequal number of rows ({len(new_df)} vs {len(data_df)})"
                    )
                if len(check_if_headers_in_dfs_are_unique(dfs=[data_df, new_df])) > 0:
                    self.remove_multiple_folders(
                        folders=[
                            self.targets_inserted_temp_path,
                            self.features_extracted_temp_path,
                            self.general_temp_folder,
                        ]
                    )
                    raise DataHeaderError(
                        msg=f"Attempted to append new feature data to features extracted file {features_path}. At least one of the new feature names aleady exist in the file."
                    )
                save_df = pd.concat([data_df, new_df], axis=1)
                write_df(df=save_df, file_type=self.file_type, save_path=data_save_path)

            if self.append_to_targets:
                print(f"Appending new data to targets inserted from {video_name}...")
                data_save_path = os.path.join(
                    self.targets_inserted_temp_path, video_name + f".{self.file_type}"
                )
                target_path = os.path.join(
                    self.targets_folder, f"{video_name}.{self.file_type}"
                )
                check_file_exist_and_readable(file_path=file_path)
                data_df = read_df(file_path=target_path, file_type=self.file_type)
                targets_clf_df = pd.DataFrame()
                if not check_same_number_of_rows_in_dfs(dfs=[data_df, new_df]):
                    self.perform_clean_up(
                        directories=[
                            self.targets_inserted_temp_path,
                            self.features_extracted_temp_path,
                            self.general_temp_folder,
                        ]
                    )
                    raise CountError(
                        f"Cannot append feature subsets to targets file of {video_name}. The files contain an unequal number of rows ({len(new_df)} vs {len(data_df)})"
                    )
                if len(check_if_headers_in_dfs_are_unique(dfs=[data_df, new_df])) > 0:
                    self.perform_clean_up(
                        directories=[
                            self.targets_inserted_temp_path,
                            self.features_extracted_temp_path,
                            self.general_temp_folder,
                        ]
                    )
                    raise DataHeaderError(
                        msg=f"Attempted to append new feature data to targets file of {target_path}. At least one of the nes feature names aleady exist in the file."
                    )
                for clf in self.clf_names:
                    if clf in data_df.columns:
                        targets_clf_df[clf] = data_df[clf]
                        data_df.drop(clf, axis=1, inplace=True)
                save_df = pd.concat([data_df, new_df, targets_clf_df], axis=1)
                write_df(df=save_df, file_type=self.file_type, save_path=data_save_path)

        if self.checks_files:
            if self.append_to_features:
                print('Confirming integrity of new "features_extracted" data...')
                file_paths = glob.glob(
                    self.features_extracted_temp_path + f"/*.{self.file_type}"
                )
                self.data_df = self.read_all_files_in_folder_mp_futures(
                    file_paths=file_paths, file_type=self.file_type
                )
                self.check_raw_dataset_integrity(
                    df=self.data_df, logs_path=self.logs_path
                )
                del self.data_df

            if self.append_to_targets:
                print('Confirming integrity of new "targets_inserted" data...')
                file_paths = glob.glob(
                    self.targets_inserted_temp_path + f"/*.{self.file_type}"
                )
                self.data_df = self.read_all_files_in_folder_mp_futures(
                    file_paths=file_paths, file_type=self.file_type
                )
                self.check_raw_dataset_integrity(
                    df=self.data_df, logs_path=self.logs_path
                )
                del self.data_df

    def replace_files_in_folders(self):
        if self.append_to_features:
            prior_to_features_append_dir = os.path.join(
                self.features_dir, f"Prior_to_feature_subset_append_{self.datetime}"
            )
            print(
                f'Replacing data in "features_extracted" directory with additional data (For safety, previous files are copied to {prior_to_features_append_dir})...'
            )
            if not os.path.isdir(prior_to_features_append_dir):
                os.makedirs(prior_to_features_append_dir)
            for file_path in glob.glob(self.features_dir + f"/*.{self.file_type}"):
                shutil.move(
                    file_path,
                    os.path.join(
                        prior_to_features_append_dir, os.path.basename(file_path)
                    ),
                )
            for file_path in glob.glob(
                self.features_extracted_temp_path + f"/*.{self.file_type}"
            ):
                shutil.move(
                    file_path,
                    os.path.join(self.features_dir, os.path.basename(file_path)),
                )

        if self.append_to_targets:
            prior_to_targets_append_dir = os.path.join(
                self.targets_folder, f"Prior_to_feature_subset_append_{self.datetime}"
            )
            print(
                f'Replacing data in "targets_inserted" directory with additional data (For safety, previous files are copied to {prior_to_targets_append_dir})...'
            )
            if not os.path.isdir(prior_to_targets_append_dir):
                os.makedirs(prior_to_targets_append_dir)
            for file_path in glob.glob(self.targets_folder + f"/*.{self.file_type}"):
                shutil.move(
                    file_path,
                    os.path.join(
                        prior_to_targets_append_dir, os.path.basename(file_path)
                    ),
                )
            for file_path in glob.glob(
                self.targets_inserted_temp_path + f"/*.{self.file_type}"
            ):
                shutil.move(
                    file_path,
                    os.path.join(self.targets_folder, os.path.basename(file_path)),
                )

        if self.save_dir != None:
            print(f"Storing new features in {self.save_dir}...")
            for file_path in glob.glob(
                self.general_temp_folder + f"/*.{self.file_type}"
            ):
                shutil.move(
                    file_path, os.path.join(self.save_dir, os.path.basename(file_path))
                )


# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_family='Frame-by-frame body-parts inside ROIs (Boolean)',
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data')
# test.run()


# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_family='Frame-by-frame body-part movements (mm)',
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data')
# test.run()

#
# test = FeatureSubsetsCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                 feature_families=['Frame-by-frame body-part distances to ROI centers (mm)', 'Frame-by-frame body-parts inside ROIs (Boolean)'],
#                                 save_dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/data',
#                                 include_file_checks=True,
#                                 append_to_features_extracted=False,
#                                 append_to_targets_inserted=True)
# test.run()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.data_processors.fsttc_calculator import FSTTCCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, Entry_Box
from simba.utils.checks import check_int
from simba.utils.enums import Defaults, Formats, Keys, Links
from simba.utils.errors import CountError


class FSTTCPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="FORWARD SPIKE TIME TILING COEFFICIENTS")
        ConfigReader.__init__(self, config_path=config_path)
        settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.FSTTC.value,
        )
        self.time_delta_eb = Entry_Box(
            settings_frm, "TIME-DELTA", "10", validation="numeric"
        )
        self.graph_cb_var = BooleanVar(value=False)
        graph_cb = Checkbutton(
            settings_frm, text="CREATE GRAPH", variable=self.graph_cb_var
        )
        self.join_bouts_within_delta_var = BooleanVar(value=False)
        join_bouts_within_delta_cb = Checkbutton(
            settings_frm,
            text="JOIN BOUTS WITHIN TIME-DELTA",
            variable=self.join_bouts_within_delta_var,
        )
        self.time_delta_at_onset_var = BooleanVar(value=False)
        time_delta_at_onset_cb = Checkbutton(
            settings_frm,
            text="TIME-DELTA AT BOUT START",
            variable=self.time_delta_at_onset_var,
        )
        settings_frm.grid(row=0, column=0, sticky=NW)
        self.time_delta_eb.grid(row=0, column=0, sticky="NW")
        join_bouts_within_delta_cb.grid(row=1, column=0, sticky="NW")
        time_delta_at_onset_cb.grid(row=2, column=0, sticky="NW")
        graph_cb.grid(row=3, column=0, sticky="NW")
        self.clf_cb_dict = self.create_cb_frame(
            main_frm=self.main_frm, cb_titles=self.clf_names, frm_title="BEHAVIORS"
        )

        self.create_run_frm(run_function=self.run, title="RUN")

        self.main_frm.mainloop()

    def run(self):
        check_int("Time delta", value=self.time_delta_eb.entry_get)
        targets = []
        for behaviour, behavior_val in self.clf_cb_dict.items():
            if behavior_val.get():
                targets.append(behaviour)

        if len(targets) < 2:
            raise CountError(
                msg="FORWARD SPIKE TIME TILING COEFFICIENTS REQUIRE 2 OR MORE BEHAVIORS.",
                source=self.__class__.__name__,
            )

        fsttc_calculator = FSTTCCalculator(
            config_path=self.config_path,
            time_window=self.time_delta_eb.entry_get,
            join_bouts_within_delta=self.join_bouts_within_delta_var.get(),
            time_delta_at_onset=self.time_delta_at_onset_var.get(),
            behavior_lst=targets,
            create_graphs=self.graph_cb_var.get(),
        )
        fsttc_calculator.run()


# _ = FSTTCPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import ast
import json
import os
from itertools import product
from typing import Dict, Union

import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_if_list_contains_values,
                                check_if_valid_input)
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Methods,
                               MLParamKeys, Options, TagNames)
from simba.utils.errors import SamplingError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import read_config_entry, write_df


class TrainMultiClassRandomForestClassifier(ConfigReader, TrainModelMixin):
    def __init__(self, config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.read_model_settings_from_config(config=self.config)

        check_if_filepath_list_is_empty(
            filepaths=self.target_file_paths,
            error_msg="Zero annotation files found in project_folder/csv/targets_inserted, cannot create model.",
        )
        print(f"Reading in {len(self.target_file_paths)} annotated files...")
        self.data_df, self.frm_idx = self.read_all_files_in_folder_mp_futures(
            annotations_file_paths=self.target_file_paths,
            file_type=self.file_type,
            classifier_names=self.clf_names,
            raise_bool_clf_error=False,
        )
        self.frm_idx = pd.DataFrame(
            {"VIDEO": list(self.data_df.index), "FRAME_IDX": self.frm_idx}
        )

        self.check_raw_dataset_integrity(df=self.data_df, logs_path=self.logs_path)
        self.data_df_wo_cords = self.drop_bp_cords(df=self.data_df)
        annotation_cols_to_remove = self.read_in_all_model_names_to_remove(
            self.config, self.clf_cnt, self.clf_name
        )
        self.x_y_df = self.delete_other_annotation_columns(
            df=self.data_df_wo_cords,
            annotations_lst=list(annotation_cols_to_remove),
            raise_error=False,
        )
        self.classifier_map = ast.literal_eval(
            read_config_entry(
                self.config,
                ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
                "classifier_map",
                data_type="str",
            )
        )
        self.under_sample_setting = (
            read_config_entry(
                self.config,
                ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
                MLParamKeys.UNDERSAMPLE_SETTING.value,
                data_type=Dtypes.STR.value,
                default_value=Dtypes.NONE.value,
            )
            .lower()
            .strip()
        )
        self.under_sample_ratio = ast.literal_eval(
            read_config_entry(
                self.config,
                ConfigKey.CREATE_ENSEMBLE_SETTINGS.value,
                MLParamKeys.UNDERSAMPLE_RATIO.value,
                data_type=Dtypes.STR.value,
            )
        )
        self.x_df, self.y_df = self.split_df_to_x_y(self.x_y_df, self.clf_name)
        check_if_list_contains_values(
            data=list(self.y_df),
            values=list(self.classifier_map.keys()),
            name=self.clf_name,
        )

        self.feature_names = self.x_df.columns
        self.check_sampled_dataset_integrity(x_df=self.x_df, y_df=self.y_df)
        print(f"Number of features in dataset: {len(self.feature_names)}")
        for k, v in self.classifier_map.items():
            print(
                f"Number of {v} frames in dataset: {len(self.y_df[self.y_df == k])} ({(len(self.y_df[self.y_df == k]) / len(self.x_df)) * 100}%)"
            )

    def _check_presence_of_classes_post_sampling(self):
        for set, clf_code in product(
            [self.y_train, self.y_test], self.classifier_map.keys()
        ):
            obs_cnt = len(self.y_df[self.y_df == clf_code])
            if obs_cnt == 0:
                raise SamplingError(
                    msg=f"Zero observations of {self.classifier_map[clf_code]} found in the training and/or test set."
                )

    def _perform_sampling(self):
        if self.split_type == Methods.SPLIT_TYPE_FRAMES.value:
            self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(
                self.x_df, self.y_df, test_size=self.tt_size
            )
        if self.under_sample_setting.lower() != Dtypes.NONE.value.lower():
            if self.under_sample_setting == "random undersample multiclass frames":
                self.x_train, self.y_train = self.random_multiclass_frm_sampler(
                    x_df=self.x_train,
                    y_df=self.y_train,
                    target_field=self.clf_name,
                    target_var=self.under_sample_ratio["target_var"],
                    sampling_ratio=self.under_sample_ratio["sampling_ratio"],
                )

            elif self.under_sample_setting == "random undersample multiclass bouts":
                self.x_train, self.y_train = self.random_multiclass_bout_sampler(
                    x_df=self.x_train,
                    y_df=self.y_train,
                    target_field=self.clf_name,
                    target_var=self.under_sample_ratio["target_var"],
                    sampling_ratio=self.under_sample_ratio["sampling_ratio"],
                )
            else:
                raise SamplingError(
                    msg=f"Under sample setting {self.under_sample_setting} not recognized. Options: [None, random undersample multiclass frames, random undersample multiclass bouts]",
                    source=self.__class__.__name__,
                )

        if self.save_train_test_frm_info:
            train_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_train.index)
            ].set_index("VIDEO")
            test_data = self.frm_idx[
                self.frm_idx.index.isin(self.x_test.index)
            ].set_index("VIDEO")
            write_df(
                df=train_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(self.eval_out_path, "train_idx.csv"),
            )
            write_df(
                df=test_data,
                file_type=Formats.CSV.value,
                save_path=os.path.join(self.eval_out_path, "test_idx.csv"),
            )
            print(
                f"Frame index of train and test set saved in {self.eval_out_path} directory..."
            )

    def run(self):
        print("Training and evaluating model...")
        self._perform_sampling()
        self.timer = SimbaTimer(start=True)
        self.rf_clf = RandomForestClassifier(
            n_estimators=self.n_estimators,
            max_depth=self.rf_max_depth,
            max_features=self.max_features,
            n_jobs=-1,
            criterion=self.criterion,
            min_samples_leaf=self.min_sample_leaf,
            bootstrap=True,
            verbose=1,
            class_weight=self.class_weights,
        )
        self.rf_clf = self.clf_fit(clf=self.rf_clf, x_df=self.x_df, y_df=self.y_df)
        if self.compute_permutation_importance in Options.PERFORM_FLAGS.value:
            self.calc_permutation_importance(
                self.x_test,
                self.y_test,
                self.rf_clf,
                self.feature_names,
                self.clf_name,
                self.eval_out_path,
            )
        if self.generate_learning_curve in Options.PERFORM_FLAGS.value:
            self.calc_learning_curve(
                x_y_df=self.x_y_df,
                clf_name=self.clf_name,
                shuffle_splits=self.shuffle_splits,
                dataset_splits=self.dataset_splits,
                tt_size=self.tt_size,
                rf_clf=self.rf_clf,
                save_dir=self.eval_out_path,
                multiclass=True,
            )
        if self.generate_precision_recall_curve in Options.PERFORM_FLAGS.value:
            self.calc_pr_curve(
                rf_clf=self.rf_clf,
                x_df=self.x_test,
                y_df=self.y_test,
                clf_name=self.clf_name,
                save_dir=self.eval_out_path,
                multiclass=True,
                classifier_map=self.classifier_map,
            )

        if self.generate_example_decision_tree in Options.PERFORM_FLAGS.value:
            self.create_example_dt(
                self.rf_clf,
                self.clf_name,
                self.feature_names,
                list(self.classifier_map.values()),
                self.eval_out_path,
            )
        if self.generate_classification_report in Options.PERFORM_FLAGS.value:
            self.create_clf_report(
                self.rf_clf,
                self.x_test,
                self.y_test,
                list(self.classifier_map.values()),
                self.eval_out_path,
                clf_name=self.clf_name,
            )
        if self.generate_features_importance_log in Options.PERFORM_FLAGS.value:
            self.create_x_importance_log(
                self.rf_clf, self.feature_names, self.clf_name, self.eval_out_path
            )
        if self.generate_features_importance_bar_graph in Options.PERFORM_FLAGS.value:
            self.create_x_importance_bar_chart(
                self.rf_clf,
                self.feature_names,
                self.clf_name,
                self.eval_out_path,
                self.feature_importance_bars,
            )
        if self.generate_example_decision_tree_fancy in Options.PERFORM_FLAGS.value:
            self.dviz_classification_visualization(
                self.x_train,
                self.y_train,
                self.clf_name,
                list(self.classifier_map.values()),
                self.eval_out_path,
            )

        if self.generate_shap_scores in Options.PERFORM_FLAGS.value:
            if not self.shap_multiprocess in Options.PERFORM_FLAGS.value:
                self.create_shap_log(
                    ini_file_path=self.config_path,
                    rf_clf=self.rf_clf,
                    x_df=self.x_train,
                    y_df=self.y_train,
                    x_names=self.feature_names,
                    clf_name=self.clf_name,
                    cnt_present=self.shap_target_present_cnt,
                    cnt_absent=self.shap_target_absent_cnt,
                    save_it=self.shap_save_n,
                    save_path=self.eval_out_path,
                )
            else:
                self.create_shap_log_mp(
                    ini_file_path=self.config_path,
                    rf_clf=self.rf_clf,
                    x_df=self.x_train,
                    y_df=self.y_train,
                    x_names=self.feature_names,
                    clf_name=self.clf_name,
                    cnt_present=self.shap_target_present_cnt,
                    cnt_absent=self.shap_target_absent_cnt,
                    save_path=self.eval_out_path,
                )

            if self.compute_partial_dependency in Options.PERFORM_FLAGS.value:
                self.partial_dependence_calculator(
                    clf=self.rf_clf,
                    x_df=self.x_train,
                    clf_name=self.clf_name,
                    save_dir=self.eval_out_path,
                )

            if self.save_meta_data in Options.PERFORM_FLAGS.value:
                meta_data_lst = [
                    self.clf_name,
                    self.criterion,
                    self.max_features,
                    self.min_sample_leaf,
                    self.n_estimators,
                    self.compute_permutation_importance,
                    self.generate_classification_report,
                    self.generate_example_decision_tree,
                    self.generate_features_importance_bar_graph,
                    self.generate_features_importance_log,
                    self.generate_precision_recall_curve,
                    self.save_meta_data,
                    self.generate_learning_curve,
                    self.dataset_splits,
                    self.shuffle_splits,
                    self.feature_importance_bars,
                    self.over_sample_ratio,
                    self.over_sample_setting,
                    self.tt_size,
                    self.split_type,
                    self.under_sample_ratio,
                    self.under_sample_setting,
                    str(self.class_weights),
                    self.rf_max_depth,
                ]

                self.create_meta_data_csv_training_one_model(
                    meta_data_lst, self.clf_name, self.eval_out_path
                )

    def save_model(self) -> None:
        """
        Method for saving pickled RF model. The model is saved in the `models/generated_models` directory of the SimBA project tree.
        """
        self.timer.stop_timer()
        self.save_rf_model(self.rf_clf, self.clf_name, self.model_dir_out)
        stdout_success(
            msg=f"Classifier {self.clf_name} saved in models/generated_models directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )
        stdout_success(
            msg=f"Evaluation files are in models/generated_models/model_evaluations folders",
            source=self.__class__.__name__,
        )


# model_trainer = TrainMultiClassRandomForestClassifier(config_path='/Users/simon/Desktop/envs/troubleshooting/multilabel/project_folder/project_config.ini')
# model_trainer.run()
# model_trainer.save_model()



import glob
import os

import h5py
import numpy as np
import pandas as pd
import scipy.io as sio

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.errors import CountError, NoFilesFoundError
from simba.utils.read_write import (find_all_videos_in_project,
                                    find_video_of_file, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_config_file)
from simba.utils.warnings import InvalidValueWarning


class APTImporterTRK(ConfigReader, PoseImporterMixin):
    def __init__(
        self,
        config_path: str,
        data_folder: str,
        id_lst: list,
        interpolation_settings: str,
        smoothing_settings: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        check_if_dir_exists(in_dir=data_folder)
        self.interpolation_settings, self.smoothing_settings = (
            interpolation_settings,
            smoothing_settings,
        )
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(
            self.logs_path, f"data_import_log_{self.datetime}.csv"
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(
            dir=self.data_folder, extensions=[".trk"]
        )
        self.data_and_videos_lk = self.link_video_paths_to_data_paths(
            data_paths=self.input_data_paths, video_paths=self.video_paths
        )
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def run(self):
        for file_cnt, (video_name, video_data) in enumerate(
            self.data_and_videos_lk.items()
        ):
            data = self.read_apt_trk_file(file_path=video_data["DATA"])
            data.columns = self.bp_headers
            data = self.intertwine_probability_cols(data=data)
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(
                    animal_bp_dict=self.animal_bp_dict,
                    video_info=get_video_meta_data(video_data["VIDEO"]),
                    video_path=video_data["VIDEO"],
                    data_df=data,
                )
                self.multianimal_identification()
                # TODO WHEN SOMEONE SHARE APT DATA.

    #
    # def import_trk(self):
    #     for file_cnt, file_path in enumerate(self.data_paths):
    #         _, file_name, file_ext = get_fn_ext(file_path)
    #         if self.animal_cnt > 0:
    #             pass
    #         video_path = find_video_of_file(self.video_dir, file_name)
    #         if not video_path:
    #             raise NoFilesFoundError(msg='Could not find a video jj')
    #         video_meta_data = get_video_meta_data(video_path=video_path)
    #         animal_tracks = self.trk_read(file_path=file_path)
    #
    #         if self.animal_cnt != 1:
    #             animal_df_lst = []
    #             for animal in animal_tracks:
    #                 m, n, r = animal.shape
    #                 out_arr = np.column_stack((np.repeat(np.arange(m), n), animal.reshape(m * n, -1)))
    #                 animal_df_lst.append(pd.DataFrame(out_arr).T.iloc[1:].reset_index(drop=True))
    #             self.animal_df = pd.concat(animal_df_lst, axis=1).fillna(0)
    #         else:
    #             m, n, r = animal_tracks.shape
    #             out_arr = np.column_stack((np.repeat(np.arange(m), n), animal_tracks.reshape(m * n, -1)))
    #             self.animal_df = pd.DataFrame(out_arr).T.iloc[1:].reset_index(drop=True)
    #         p_cols = pd.DataFrame(1, index=self.animal_df.index, columns=self.animal_df.columns[1::2] + .5)
    #         self.animal_df = pd.concat([self.animal_df, p_cols], axis=1).sort_index(axis=1)
    #         if len(self.bp_headers) != len(self.animal_df.columns):
    #             raise CountError(msg=f'SimBA detected {str(len(self.animal_df.columns))} body-parts in the .TRK file {file_path}. Your SimBA project however expects {len(self.bp_headers)} body-parts')
    #
    #
    #         max_dim = max(video_meta_data['width'], video_meta_data['height'])
    #         self.circle_scale = int(self.radius_scaler / (self.resolution_scaler / max_dim))
    #         self.font_scale = float(self.font_scaler / (self.resolution_scaler / max_dim))
    #         self.spacingScale = int(self.space_scaler / (self.resolution_scaler / max_dim))
    #         cv2.namedWindow('Define animal IDs', cv2.WINDOW_NORMAL)
    #         self.cap = cv2.VideoCapture(video_path)
    #         self.create_first_interface()
    #
    # def __insert_all_animal_bps(self, frame=None):
    #     for animal, bp_data in self.img_bp_cords_dict.items():
    #         for bp_cnt, bp_tuple in enumerate(bp_data):
    #             try:
    #                 cv2.circle(frame, bp_tuple, self.circle_scale, self.animal_bp_dict[animal]['colors'][bp_cnt], -1, lineType=cv2.LINE_AA)
    #             except Exception as err:
    #                 if type(err) == OverflowError:
    #                     InvalidValueWarning(f'SimBA encountered a pose-estimated body-part located at pixel position {str(bp_tuple)}. This value is too large to be converted to an integer. Please check your pose-estimation data to make sure that it is accurate.')
    #                 print(err.args)
    #
    # def create_first_interface(self):
    #     while True:
    #         self.cap.set(1, self.frm_number)
    #         _, self.frame = self.cap.read()
    #         self.overlay = self.frame.copy()
    #         self.img_bp_cords_dict = {}
    #         for animal_cnt, (animal_name, animal_bps) in enumerate(self.animal_bp_dict.items()):
    #             self.img_bp_cords_dict[animal_name] = []
    #             for bp_cnt in range(len(animal_bps['X_bps'])):
    #                 x_cord = int(self.animal_df.loc[self.frm_number, animal_name + '_' + animal_bps['X_bps'][bp_cnt]])
    #                 y_cord = int(self.animal_df.loc[self.frm_number, animal_name + '_' + animal_bps['Y_bps'][bp_cnt]])
    #                 self.img_bp_cords_dict[animal_name].append((x_cord, y_cord))
    #         self.__insert_all_animal_bps(frame=self.overlay)


test = APTImporterTRK(
    config_path="/Users/simon/Desktop/envs/troubleshooting/trk_test/project_folder/project_config.ini",
    data_folder="/Users/simon/Desktop/envs/troubleshooting/trk_test/data/",
    id_lst=["Animal_1", "Animal_2", "Animal_3"],
    interpolation_settings="Body-parts: Nearest",
    smoothing_settings={
        "Method": "Savitzky Golay",
        "Parameters": {"Time_window": "200"},
    },
)
test.run()


# test = TRKImporter(config_path='/Users/simon/Desktop/envs/troubleshooting/DLC_2_Black_animals/project_folder/project_config.ini',
#                    data_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/trk_data',
#                    animal_id_lst=['Animal_1', 'Animal_2'],
#                    interpolation_method="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})


# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Hornet/project_folder/project_config.ini',
#                  data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Hornet_single_slp/import',
#                  actor_IDs=['Hornet'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()


# def __init__(self,
#              config_path: str,
#              data_folder: str,
#              animal_id_lst: list,
#              interpolation_method: str,
#              smooth_settings: dict):



__author__ = "Simon Nilsson"

import os
from copy import deepcopy

import numpy as np
import pandas as pd

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import check_that_column_exist
from simba.utils.enums import Methods, TagNames
from simba.utils.errors import CountError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (clean_sleap_file_name,
                                    find_all_videos_in_project, get_fn_ext,
                                    get_video_meta_data, write_df)

TRACK = "track"
INSTANCE_SCORE = "instance.score"


class SLEAPImporterCSV(ConfigReader, PoseImporterMixin):
    """
    Importing SLEAP pose-estimation data into SimBA project in ``CSV`` format.

    .. note::
      `Google Colab notebook for converting SLEAP .slp to CSV written by @Toshea111  <https://colab.research.google.com/drive/1EpyTKFHVMCqcb9Lj9vjMrriyaG9SvrPO?usp=sharing>`__.
      `Example expected SLEAP csv data file for 5 animals / 4 pose-estimated body-parts  <https://github.com/sgoldenlab/simba/blob/master/misc/sleap_csv_example.csv>`__.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str data_folder: Path to folder containing SLEAP data in `csv` format.
    :parameter List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :parameter str interpolation_settings: String defining the pose-estimation interpolation method. OPTIONS: 'None', 'Animal(s): Nearest',
        'Animal(s): Linear', 'Animal(s): Quadratic','Body-parts: Nearest', 'Body-parts: Linear',
        'Body-parts: Quadratic'.
    :parameter str smoothing_settings: Dictionary defining the pose estimation smoothing method. EXAMPLE: {'Method': 'Savitzky Golay',
        'Parameters': {'Time_window': '200'}})

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`,
           2022.

    >>> sleap_csv_importer = SLEAPImporterCSV(config_path=r'project_folder/project_config.ini', data_folder=r'data_folder', actor_IDs=['Termite_1', 'Termite_2', 'Termite_3', 'Termite_4', 'Termite_5'], interpolation_settings="Body-parts: Nearest", smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
    >>> sleap_csv_importer.run()
    """

    def __init__(
        self,
        config_path: str,
        data_folder: str,
        id_lst: list,
        interpolation_settings: str,
        smoothing_settings: dict,
    ):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        self.interpolation_settings, self.smoothing_settings = (
            interpolation_settings,
            smoothing_settings,
        )
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(
            self.logs_path, f"data_import_log_{self.datetime}.csv"
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.input_data_paths = self.find_data_files(
            dir=self.data_folder, extensions=[".csv"]
        )
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.data_and_videos_lk = self.link_video_paths_to_data_paths(
                data_paths=self.input_data_paths,
                video_paths=self.video_paths,
                filename_cleaning_func=clean_sleap_file_name,
            )
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(
                self.multi_animal_status,
                self.id_lst,
                self.animal_cnt,
                self.x_cols,
                self.y_cols,
                self.p_cols,
                self.clr_lst,
            )
            self.update_bp_headers_file()
        else:
            self.data_and_videos_lk = dict(
                [
                    (get_fn_ext(file_path)[1], {"DATA": file_path, "VIDEO": None})
                    for file_path in self.input_data_paths
                ]
            )
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def run(self):
        for file_cnt, (video_name, video_data) in enumerate(
            self.data_and_videos_lk.items()
        ):
            output_filename = clean_sleap_file_name(filename=video_name)
            print(f"Analysing {output_filename}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = video_name
            self.save_path = os.path.join(
                os.path.join(self.input_csv_dir, f"{output_filename}.{self.file_type}")
            )
            data_df = pd.read_csv(video_data["DATA"])
            if INSTANCE_SCORE in data_df.columns:
                data_df = data_df.drop([INSTANCE_SCORE], axis=1)
            idx = data_df.iloc[:, :2]
            check_that_column_exist(df=idx, column_name=TRACK, file_name=video_name)
            idx[TRACK] = idx[TRACK].fillna("track_1")
            idx[TRACK] = idx[TRACK].str.replace(r"[^\d.]+", "").astype(int)
            data_df = data_df.iloc[:, 2:].fillna(0)
            if self.animal_cnt > 1:
                self.data_df = pd.DataFrame(
                    self.transpose_multi_animal_table(
                        data=data_df.values, idx=idx.values, animal_cnt=self.animal_cnt
                    )
                )
            else:
                idx = list(idx.drop(TRACK, axis=1)["frame_idx"])
                self.data_df = data_df.set_index([idx]).sort_index()
                self.data_df.columns = np.arange(len(self.data_df.columns))
                self.data_df = self.data_df.reindex(
                    range(0, self.data_df.index[-1] + 1), fill_value=0
                )

            if len(self.bp_headers) != len(self.data_df.columns):
                raise CountError(
                    msg=f"SimBA project expects {len(self.bp_headers)} data columns, but your SLEAP data file {video_name} contains {len(self.data_df.columns)} columns. Missing columns: {list(set(self.bp_headers) - set(self.data_df.columns))}",
                    source=self.__class__.__name__,
                )
            self.data_df.columns = self.bp_headers
            self.out_df = deepcopy(self.data_df)
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(
                    animal_bp_dict=self.animal_bp_dict,
                    video_info=get_video_meta_data(video_data["VIDEO"]),
                    data_df=self.data_df,
                    video_path=video_data["VIDEO"],
                )
                self.multianimal_identification()
            write_df(
                df=self.out_df,
                file_type=self.file_type,
                save_path=self.save_path,
                multi_idx_header=True,
            )
            if self.interpolation_settings != "None":
                self.__run_interpolation()
            if self.smoothing_settings["Method"] != "None":
                self.__run_smoothing()
            video_timer.stop_timer()
            stdout_success(
                msg=f"Video {video_name} data imported...",
                elapsed_time=video_timer.elapsed_time_str,
                source=self.__class__.__name__,
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"{len(list(self.data_and_videos_lk.keys()))} file(s) imported to the SimBA project (project_folder/csv/input_csv directory)",
            source=self.__class__.__name__,
        )

    def __run_interpolation(self):
        print(
            f"Interpolating missing values in video {self.video_name} (Method: {self.interpolation_settings})..."
        )
        _ = Interpolate(
            input_path=self.save_path,
            config_path=self.config_path,
            method=self.interpolation_settings,
            initial_import_multi_index=True,
        )

    def __run_smoothing(self):
        print(
            f'Performing {self.smoothing_settings["Method"]} smoothing on video {self.video_name}...'
        )
        Smooth(
            config_path=self.config_path,
            input_path=self.save_path,
            time_window=int(self.smoothing_settings["Parameters"]["Time_window"]),
            smoothing_method=self.smoothing_settings["Method"],
            initial_import_multi_index=True,
        )


# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Hornet/project_folder/project_config.ini',
#                  data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Hornet_single_slp/import',
#                  id_lst=['Hornet'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Hornet/project_folder/project_config.ini',
#                  data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Hornet_single_slp/import',
#                  id_lst=['Hornet'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterCSV(config_path=r'/Users/simon/Desktop/envs/troubleshooting/slp_1_animal_1_bp/project_folder',
#                  data_folder='/Users/simon/Desktop/envs/troubleshooting/slp_1_animal_1_bp/import',
#                  id_lst=['Termite_1'],
#                  interpolation_settings="Body-parts: Nearest",
#                  smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()
